"
 COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass: #BinaryObjectStorage
	instanceVariableNames: 'manager stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-BinaryStorage'
!

!BinaryObjectStorage class methodsFor: 'documentation'!

copyright
"
 COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /claus/smalltalk/st/goodies/Persistency/RCS/BOSS.st,v 1.5 95/02/18 20:33:31 claus Exp $
"
!

documentation
"
    This class provides an ST-80 compatible (I hope) interface to
    BinaryInput- and BinaryOutputManager.
    The protocol is not complete, PP seems to offer more (skip etc.)
    this will be added as time goes by ...

    The protocol has been implemented by 'guessing' what users of this
    class expect; no warranty here. I depend on your feedback in adding
    more for compatibility; let me know what is required ....
"
!

examples 
"
    example (taken from an article in comp.lang.smalltalk):

      storing:
	|bos testArray |

	testArray := Array new: 4.
	1 to: 4 do: 
	    [:index |
	    | innerarray |
	    innerarray := Array new: 10.
	    1 to: 10 do:
		    [:innerIndex |
		    | innerinnerArray |
		    innerinnerArray := Array new: 1000 withAll: 0.
		    innerarray at: innerIndex put: innerinnerArray].
	    testArray at: index put: innerarray
	].

	bos := BinaryObjectStorage onNew: (Filename named: 'array.boss') writeStream.
	bos nextPut: testArray.
	bos nextPut: #(1 2 3 4).
	bos nextPut: 'hello'.
	bos close.

      retrieving:

	|bos obj1 obj2 obj3|
	bos := BinaryObjectStorage onOld: (Filename named: 'array.boss') readStream.
	obj1 := bos next.
	obj2 := bos next.
	obj3 := bos next.
	bos close.
	obj1 inspect.
	obj2 inspect.
	obj3 inspect.
"
! !

!BinaryObjectStorage class methodsFor:'signal access'!

binaryLoadErrorSignal  
    "return the signal raised, when a binary load fails due to format errors,
     for example, if you try to fileIn a binary object from another smalltalk,
     or from a stream which does not contain binary objects at all."

    ^ BinaryIOManager binaryLoadErrorSignal 
!

nonexistingClassSignal
    "return the signal raised, when an instance of a class is about to be filed in,
     which does not (or no-longer) exist."

    ^ BinaryIOManager nonexistingClassSignal
!

changedInstSizeSignal
    "return the signal raised, when an instance of a class is about to be filed in,
     which has a different inst-size (i.e. the number of named instvars has changed)"

    ^ BinaryIOManager changedInstSizeSignal
!

changedInstLayoutSignal
    "return the signal raised, when an instance of a class is about to be filed in,
     which has a different inst layout (i.e. the size matches, but the order of
     instvars has changed)"

    ^ BinaryIOManager changedInstLayoutSignal
!

changedIndexedInstSignal
    "return the signal raised, when an instance of a class is about to be filed in,
     which has different indexed instvars (i.e. byte-insts vs. pointers etc)"

    ^ BinaryIOManager changedIndexedInstSignal
! !

!BinaryObjectStorage class methodsFor: 'instance creation'!

onOld:aStream
    "return an instance of myself to read objects from aStream"

    ^ self new onOld:aStream
!

onNew:aStream
    "return an instance of myself to write objects to aStream"

    ^ self new onNew:aStream
! !

!BinaryObjectStorage methodsFor: 'private'!

onOld:aStream
    stream := aStream.
    stream binary.
    manager := BinaryInputManager new:1024.
!

onNew:aStream
    stream := aStream.
    stream binary.
    manager := (BinaryOutputManager new:1024) initialize.
! !

!BinaryObjectStorage methodsFor: 'object io'!

next
    "read and return the next object"

    ^ manager readFrom:stream
!

nextPut:anObject
    "write an object"

    anObject storeBinaryOn: stream manager: manager
!

close
    stream close
! !
