"
 COPYRIGHT (c) 1994 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Block subclass:#CheapBlock
       instanceVariableNames:'selfValue method'
       classVariableNames:''
       poolDictionaries:''
       category:'Kernel-Methods'
!

CheapBlock comment:'
COPYRIGHT (c) 1994 by Claus Gittinger
	      All Rights Reserved

$Header: /claus/smalltalk/st/libbasic/RCS/CheapBlk.st,v 1.4 95/02/05 21:28:39 claus Exp $
'!

!CheapBlock class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1994 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /claus/smalltalk/st/libbasic/RCS/CheapBlk.st,v 1.4 95/02/05 21:28:39 claus Exp $
"
!

documentation
"
    CheapBlocks are blocks which do not need their home-context
    (i.e. blocks that do not access any method arguments or method locals).
    Since they have no reference to the home, they must store their
    creating method explicitely - otherwise, the system had no chance of
    finding the source-position of the block.

    Instance variables:

      selfValue   <Object>          value to use for self if its a copying block

      method      <Method>          method where block was created 

    NOTICE: layout known by runtime system and compiler - do not change
"
! !

!CheapBlock class methodsFor:'instance creation'!

code:codeAddress byteCode:bCode numArgs:numArgs sourcePosition:sourcePos initialPC:initialPC literals:literals dynamic:dynamic
    "create a new cheap (homeless) block.
     Not for public use - special hook for the compiler."

    |newBlock|

    newBlock := super basicNew code:codeAddress 
			   byteCode:bCode
			   numArgs:numArgs
		     sourcePosition:sourcePos
			  initialPC:initialPC
			   literals:literals
			    dynamic:dynamic.
    ^ newBlock
! !

!CheapBlock methodsFor:'private accessing'!

code:codeAddress byteCode:bCode numArgs:numArgs sourcePosition:srcPos initialPC:iPC literals:lits dynamic:dynamic
    "set all relevant internals"

    self code:codeAddress.
    byteCode := bCode.
    nargs := numArgs.
    sourcePos := srcPos.
    initialPC := iPC.
    literals := lits.
    self dynamic:dynamic
! !

!CheapBlock methodsFor:'accessing'!

selfValue
    "return the copied self"

    ^ selfValue
!

method
    "return the receivers home method.
     Thats the method where the block was created."

    ^ method
! !

!CheapBlock methodsFor:'printing & storing'!

printOn:aStream
    "append a a printed representation of the block to aStream"

    |class selector pair|

    "
     cheap blocks have no home context, but a method instead
    "

    aStream nextPutAll:'[] in '.
    "
     find out, for which class this method was for ...
    "
    method notNil ifTrue:[
	pair := method who.
	pair notNil ifTrue:[
	    class := pair at:1.
	    selector := pair at:2.
	    aStream nextPutAll:(class name , '-' , selector).
	    aStream nextPutAll:' (optimized)'.
	    ^ self
	].
    ].
    "
     currently, some cheap blocks don't know where they have been created
    "
    aStream nextPutAll:' ??? (optimized)'.
    ^ self
! !

