TreeView subclass:#ClassTreeView
       instanceVariableNames:'subclassInfo'
       classVariableNames:''
       poolDictionaries:''
       category:'Demos-Graphics'
!

ClassTreeView comment:'
 a small application showing an ObjectView / TreeView use

 must fileIn TreeView.st first,
 then start with:
     ClassTreeView open

 %W% %E%
 written spring 91 by claus
'!

!ClassTreeView class methodsFor:'startup'!

open
    |aDemo topView v|

    topView := StandardSystemView 
		    label:'Class hierarchy'
		    icon:(Form fromFile:'Demo.icon' resolution:100).
    v  := HVScrollableView for:self in:topView.
    v origin:(0 @ 0) extent:(1.0 @ 1.0).

    topView open

    "ClassTreeView open"
! !

!ClassTreeView methodsFor:'initialization'!

initialize
    topNode := Object.
    super initialize.
!

initializeMiddleButtonMenu
    |labels|

    labels := resources array:#(
			'browse'
			'browse hierarchy'
			'-'
			'category'
			'comment'
			'definition'
		       ).

    self middleButtonMenu:(PopUpMenu
				labels:labels
			     selectors:#(startBrowser
					 startHierarchyBrowser
					 nil
					 showCategory
					 showComment
					 showDefinition
					)
				receiver:self
				     for:self)
!

updateTree
    |d|

    d := IdentityDictionary new.
    Smalltalk allBehaviorsDo:[:aClass |
	|superClass set|

	superClass := aClass superclass.
	superClass notNil ifTrue:[
	    set := d at:superClass ifAbsent:[].
	    set isNil ifTrue:[
		d at:superClass put:(IdentitySet with:aClass)
	    ] ifFalse:[
		set add:aClass.
	    ].
	]
    ].
    subclassInfo := d.
    super updateTree
! !

!ClassTreeView methodsFor:'generating the class-tree picture'!

childrenOf:aClass do:aBlock
    |subclasses subclassNames|

    subclasses := subclassInfo at:aClass ifAbsent:nil.
    "/ subclasses := aClass subclasses.

    (subclasses size ~~ 0) ifTrue:[
	subclasses := subclasses asSortedCollection:[:cls1 :cls2 | cls1 name < cls2 name].
	subclasses do:[:aClass |
	    aBlock value:aClass
	]
    ]
!

textFor:aClass
    ^ aClass name
! !

!ClassTreeView methodsFor:'user interaction'!

startBrowser
    self withSelectedElementDo:[:theClass |
	 SystemBrowser browseClass:theClass
    ]
!

startHierarchyBrowser
    self withSelectedElementDo:[:theClass |
	 SystemBrowser browseClassHierarchy:theClass
    ]
!

showCategory
    self withSelectedElementDo:[:theClass |
	 self notify:(theClass category printString)
    ]
!

showComment
    self withSelectedElementDo:[:theClass |
	 self notify:(theClass comment printString)
    ]
!

showDefinition
    |s|

    self withSelectedElementDo:[:theClass |
	s := WriteStream on:String new.
	theClass fileOutDefinitionOn:s.
	self notify:s contents
    ]
! !
