"
 COPYRIGHT (c) 1991 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

View subclass:#ClockView
	 instanceVariableNames:'lastTime timeBlock
				xTable yTable 
				showSeconds showMinutes showHours
				bgColor fgColor'
	 classVariableNames:''
	 poolDictionaries:''
	 category:'Demos'
!

!ClockView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1991 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /claus/smalltalk/st/libwidg3/RCS/ClockView.st,v 1.4 95/02/22 01:23:01 claus Exp $
"
!

documentation
"
 xclock replacement

 :-) not that I thought this clock is any better than xclock
     or that the world needs clocks ...

     its showing how easy it is to program in Smalltalk - 
     this took less than an hour to program - compare with xclock code ....

 Since this is a subclass of view, you can use it as a wdiget in
 more complex views.

 'in a topView:'

    |top clk|

    top := StandardSystemView new.
    top label:'ST clock'.
    top extent:200@200.

    clk := ClockView origin:0.0@0.0 corner:1.0@1.0 in:top.

    top open

 'as a component:'

    |top frame clk fileList|

    top := StandardSystemView new.
    top extent:200@200.

    frame := View origin:1.0@0.0 corner:1.0@50 in:top.
    frame leftInset:-50.
    frame level:-1.

    clk := ClockView origin:0.0@0.0 corner:1.0@1.0 in:frame.
    clk showSeconds:false.

    fileList := ScrollableView for:FileSelectionList in:top.
    fileList origin:0@50 corner:1.0@1.0. 

    top open


 written spring 91 by claus
"
! !

!ClockView methodsFor:'initialization'!

initialize
    |angle ta|

    super initialize.

    bgColor := White.
    fgColor := Black.
    self viewBackground:White.

    showSeconds := showMinutes := showHours := true.

    "initialize table of (normalized) x/y coordinates of ticks"
    xTable := Array new:60.
    yTable := Array new:60.
    angle := 0.
    1 to:60 do:[:index |
	angle := angle + 6. "360 / 60"
	ta := (360 - angle) + 90.
	xTable at:index put:((ta degreesToRadians) cos).
	yTable at:index put:((ta degreesToRadians) sin negated)
    ]
!

initializeMiddleButtonMenu
    |labels|

    labels := resources array:#('seconds').
    self middleButtonMenu:(PopUpMenu
				    labels:labels
				 selectors:#(toggleSeconds)
				    receiver:self
					 for:self)
! !

!ClockView methodsFor:'misc'!

showSeconds:aBoolean
    showSeconds := aBoolean.
    shown ifTrue:[
	Processor removeTimedBlock:timeBlock.
	lastTime := nil.
	self clear.
	self redraw.
	Processor addTimedBlock:timeBlock afterSeconds:1
    ]
!

toggleSeconds
    "user wants to toggle seconds flag"

    showSeconds := showSeconds not.
    Processor removeTimedBlock:timeBlock.
    lastTime := nil.
    self clear.
    self redraw.
    Processor addTimedBlock:timeBlock afterSeconds:1
! !

!ClockView methodsFor:'drawing'!

showTime
    "executed every second (by timedBlock)"

    |now anyChange deltaT|

    shown ifFalse:[^ self].

    now := Time now.
    lastTime isNil ifTrue:[
	anyChange := true
    ] ifFalse:[
	showSeconds ifTrue:[
	    anyChange := now ~= lastTime
	] ifFalse:[
	    anyChange := now hours ~= lastTime hours.
	    anyChange ifFalse:[
		anyChange := now minutes ~= lastTime minutes
	    ]
	]
    ].
    anyChange ifTrue:[
	self updateArms.
	device synchronizeOutput
    ].
    showSeconds ifTrue:[
	deltaT := 1
    ] ifFalse:[
	deltaT := 30
    ].
    Processor addTimedBlock:timeBlock afterSeconds:deltaT
!

redraw
    lastTime isNil ifTrue:[
	lastTime := Time now
    ].
    self redrawTicks.
    self redrawArms
!

updateArms
    "clear previous arms - draw new arms"

    |hr min sec lastSec lastHr lastMin now tm ts th|

    now := Time now.
    lastTime notNil ifTrue:[
	"a little optimization - avoid redrawing minute-
	 and hour arms if second-arms distance is big enough"

	showSeconds ifTrue:[
	    lastHr := lastTime hours.
	    lastMin := lastTime minutes.
	    lastSec := lastTime seconds.

	    hr := now hours.
	    min := now minutes.
	    sec := now seconds.

	    ((min == lastMin) and:[hr == lastHr]) ifTrue:[
		"only second has not changed"
		(hr >= 12) ifTrue:[
		    hr := hr - 12
		].
		hr := hr * 5.
		hr := hr + (min * 5 / 60) asInteger.

		min < 30 ifTrue:[
		    tm := min + 60
		] ifFalse:[
		    tm := min
		].
		sec < 30 ifTrue:[
		    ts := sec + 60
		] ifFalse:[
		    ts := sec
		].
		hr < 30 ifTrue:[
		    th := hr + 60
		] ifFalse:[
		    th := hr
		].

		(((tm - ts) abs > 5) and:[(th - ts) abs > 5]) ifTrue:[
		    "second is more than 5 ticks away from others -
		     no need to draw other arms"

		    lastMin < 30 ifTrue:[
			tm := lastMin + 60
		    ] ifFalse:[
			tm := lastMin
		    ].
		    lastSec < 30 ifTrue:[
			ts := lastSec + 60
		    ] ifFalse:[
			ts := lastSec
		    ].
		    lastHr < 30 ifTrue:[
			th := lastHr + 60
		    ] ifFalse:[
			th := lastHr
		    ].
		    "if lastsecond was more than 5 away from others,
		     just draw seconds-arm"
		    (((tm - ts) abs > 5) and:[(th - ts) abs > 5]) ifTrue:[
			self paint:bgColor.
			self drawArm:lastSec width:1 len:0.8.
			lastTime := now.
			self paint:fgColor.
			self drawArm:sec width:1 len:0.8.
			^ self
		    ]
		]
	    ]
	].

	self paint:bgColor.
	self drawArmsFor:lastTime
    ].
    lastTime := now.
    self paint:fgColor.
    self drawArmsFor:lastTime
!

redrawArms
    "draw current arms"

    self drawArmsFor:lastTime
!

drawArmsFor:aTime
    "draw arms for a given time in current paint"

    |sec min hr hrTick|

    aTime isNil ifTrue:[^ self].
    hr := aTime hours.
    min := aTime minutes.
    sec := aTime seconds.

    (hr >= 12) ifTrue:[
	hr := hr - 12
    ].
    "
     adjust hour for minute value ...
    "
    hrTick := hr * 5.
    hrTick := hrTick + (min * 5 / 60) asInteger.

    showHours ifTrue:[
	self drawArm:hrTick width:3 len:0.5.
    ].
    showMinutes ifTrue:[
	self drawArm:min width:3 len:0.8.
    ].
    showSeconds ifTrue:[
	self drawArm:sec width:1 len:0.8
    ]
!

drawArm:tick width:w len:l
    "draw an arm of the clock; argument tick specifies minute (0..59); 
     width is width in pixel, len is relative len - 1.0 is full"

    |index p1x p1y p2x p2y|

    index := tick.
    (index == 0) ifTrue:[
	index := 60
    ].
    p1x := (width - w) // 2.
    p1y := height // 2.
    p2x := p1x + (((width / 2) * (xTable at:index)) * l) + 0.5.
    p2y := p1y + (((height / 2) * (yTable at:index)) * l) + 0.5.
    self lineWidth:w.
    self displayLineFromX:p1x y:p1y
		      toX:(p2x asInteger) y:(p2y asInteger)
!

redrawTicks
    |cx cy p1x p1y p2x p2y nx ny long drawShorts tickStart|

    self lineWidth:0.
    self paint:fgColor.

    cx := width // 2.
    cy := height // 2.
    drawShorts := (width >= 100) and:[height >= 100].

    1 to:60 do:[:tick |

	"is it a long tick ?"

	long := (tick \\ 5) == 0.

	(drawShorts or:[long]) ifTrue:[

	    "get normalized x/y of tick-end"

	    nx := (width / 2) * (xTable at:tick).
	    ny := (height / 2) * (yTable at:tick).

	    "if long, tick goes from 0.85 to 0.95;
	     otherwise from 0.90 to 0.95"

	    long ifTrue:[
		tickStart := 0.85
	    ] ifFalse:[
		tickStart := 0.9
	    ].

	    p1x := cx + (nx * tickStart).
	    p1y := cy + (ny * tickStart).
	    p2x := cx + (nx * 0.95).
	    p2y := cy + (ny * 0.95).
	    self displayLineFromX:(p1x asInteger) y:(p1y asInteger)
			      toX:(p2x asInteger) y:(p2y asInteger)
	]
    ]
! !

!ClockView methodsFor:'events'!

update:what with:arg
    (what == #rectangle) ifTrue:[
	self redraw
    ]
!

sizeChanged
    "have to clear & redraw when my size changes"

    super sizeChanged.
    self clear.
    self redraw
!

mapped
    super mapped.

    "trigger timeout block in a second"
    timeBlock := [self showTime].
    Processor addTimedBlock:timeBlock afterSeconds:2
! !

!ClockView methodsFor:'user actions'!

destroy
    "remove myself"

    Processor removeTimedBlock:timeBlock.
    super destroy
! !
