"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

DrawObject subclass:#DrawText
     instanceVariableNames:'fontFamily fontFace fontStyle fontSize font text cursorLine'
     classVariableNames:''
     poolDictionaries:''
     category:'Applications-DrawTool-Objects'
!

DrawText comment:'

COPYRIGHT (c) 1989-93 by Claus Gittinger
	      All Rights Reserved

I represent text

$Header: /claus/smalltalk/st/clients/DrawTool/RCS/DrawText.st,v 1.6 95/02/18 00:11:23 claus Exp $
written summer 89 by claus
'!

!DrawText class methodsFor:'interface to DrawView'!

needsKeyboardInput
    ^ true
!

hasFixedSize
    ^ true
! !

!DrawText class methodsFor:'instance creation'!

new
    |newDText|

    newDText := super new.
    newDText cursorLine:1.
    newDText setFontFamily:'times' face:'medium' style:'roman' size:12.
    newDText getFont.
    newDText text:(StringCollection new:1).
    ^ newDText
!

family:aFamily face:aFace style:aStyle size:aSize
    |newDText|

    newDText := super new.
    newDText cursorLine:1.
    newDText setFontFamily:aFamily face:aFace style:aStyle size:aSize.
    newDText text:(StringCollection new:1).
    ^ newDText

! !

!DrawText methodsFor:'copying'!

copy
    "answer a copy - have to copy the string here"

    |aCopy|

    aCopy := super copy.
    aCopy text:(text deepCopy).
    ^ aCopy
! !

!DrawText methodsFor:'accessing'!

cursorLine:aNumber
    cursorLine := aNumber
!

text:aText
    text := aText asStringCollection.
    self computeBoundingBox
!

setFontFamily:familyString face:faceString style:aStyleString size:aNumber
    fontFamily := familyString.
    fontFace := faceString.
    fontSize := aNumber.
    fontStyle := aStyleString.
    self getFont
!

setFontFamily:aString
    fontFamily := aString.
    self getFont
!

setFontFace:aString
    fontFace := aString.
    self getFont
!

setFontStyle:aStyleString
    fontStyle := aStyleString.
    self getFont
!

setFontSize:aNumber
    fontSize := aNumber.
    self getFont
!

fontFamily:aFamily
    fontFamily := aFamily.
    self getFont.
    self computeBoundingBox
!

fontFace:aFace
    fontFace := aFace.
    self getFont.
    self computeBoundingBox
!

fontStyle:aStyle
    fontStyle := aStyle.
    self getFont.
    self computeBoundingBox
!

fontSize:aSize
    fontSize := aSize.
    self getFont.
    self computeBoundingBox
!

fontFace
    ^ fontFace

!

fontFamily
    ^ fontFamily

!

fontStyle
    ^ fontStyle

!

fontSize
    ^ fontSize

!

fontFamily:aFamily face:aFace style:aStyle size:aSize
    self setFontFamily:aFamily face:aFace style:aStyle size:aSize.
    self computeBoundingBox

!

font:aFont
    font := aFont.
    self computeBoundingBox

!

text
    ^ text
! !

!DrawText methodsFor:'binary storage'!

readBinaryContentsFrom: stream manager: manager
    "must recreate font"

    super readBinaryContentsFrom: stream manager: manager.
    self getFont
! !

!DrawText methodsFor:'private'!

getFont
    font := Font family:fontFamily face:fontFace
		  style:fontStyle size:fontSize.
    font isNil ifTrue:[
       font := Font defaultFont
    ].
    font notNil ifTrue:[
	font := font on:Display
    ]
! !

!DrawText methodsFor:'frame computation'!

computeBoundingBox
    |maxLen thisLen|

    origin isNil ifTrue:[^ self].
    font isNil ifTrue:[
	corner := origin
    ] ifFalse:[
	maxLen := 0.
	text do:[:line |
	    line notNil ifTrue:[
		thisLen := font widthOf:line.
		(thisLen > maxLen) ifTrue:[maxLen := thisLen]
	    ]
	].
	(maxLen == 0) ifTrue:[
	    corner := origin
	] ifFalse:[
	    corner := (origin + (maxLen @ (text size * font height)))
	]
    ].
    frame := Rectangle origin:origin corner:corner
! !

!DrawText methodsFor:'editing'!

handlesKeyboardInput
    "return true, if this objects is interrested in keyboard input"

    ^ true
!

keyInput:key
    |line|

    (key == #Return) ifTrue: [
	 text grow:(text size + 1).
	 cursorLine := cursorLine + 1
    ] ifFalse: [
	line := text at:cursorLine.
	(key == #BackSpace) ifTrue:[
	    line isNil ifFalse:[
		line size > 0 ifTrue:[
		    text at:cursorLine put:(line copyFrom:1 to:(line size - 1))
		]
	    ]
	] ifFalse:[
	    (key isMemberOf:Character) ifFalse:[^ self].
	    line isNil ifTrue:[
		text at:cursorLine put:(key asString)
	    ] ifFalse:[
		text at:cursorLine put:(line , (key asString))
	    ]
	]
    ].
    self computeBoundingBox
! !

!DrawText methodsFor:'drawing'!

drawIn:aGC offset:anOffset
    |r org|

    org := origin + anOffset.

    foreground isNil ifTrue:[
	foreground := Color black.
	background := Color white
    ].

    fillPattern isNil ifTrue:[
	"no fill"
    ] ifFalse:[
	(fillPattern == 1) ifTrue:[
	    "solid background"
	    "linePattern is also solid, it can be done faster"
	    linePattern == 1 ifTrue:[
		aGC paint:foreground on:background.
		self drawTextIn:aGC x:(org x) y:(org y) opaque:true.
		^ self
	    ] ifFalse:[
		r := Rectangle origin:org corner:(corner + anOffset).
		aGC paint:background.
		aGC fillRectangle:r
	    ]
	] ifFalse:[
	    "a real pattern"
	    aGC setMaskOriginX:org x y:org y.
	    r := Rectangle origin:org corner:(corner + anOffset).
	    aGC paint:foreground on:background.
	    aGC withPattern:fillPattern do:[
	       aGC fillRectangle:r
	    ]
	]
    ].

    linePattern isNil ifTrue:[
	"no drawing"
    ] ifFalse:[
	(linePattern == 1) ifTrue:[
	    "solid foreground"
	    aGC paint:foreground on:background.
	    self drawTextIn:aGC x:(org x) y:(org y) opaque:false
	] ifFalse:[
	    "a real pattern"
	    aGC setMaskOriginX:org x y:org y.
	    aGC paint:foreground on:background.
	    aGC withPattern:linePattern do:[
		self drawTextIn:aGC x:(org x) y:(org y) opaque:false
	    ]
	]
    ]
!

drawOutlineIn:aGC offset:anOffset
    |org x y|

    org := origin + anOffset.
    x := org x.
    y := org y.
    fillPattern notNil ifTrue:[
	 aGC displayRectangle:(Rectangle
				 origin:org
				 corner:(corner + anOffset))
    ].
    self drawTextIn:aGC x:x y:y opaque:false
!

drawTextIn:aGC x:x y:y opaque:opaque
    |ypos delta|

    text notNil ifTrue:[
	ypos := y + font ascent.
	delta := font height.
	aGC font:font.
	text do:[:line |
	    line notNil ifTrue:[
		opaque ifTrue:[
		    aGC displayOpaqueString:line x:x y:ypos
		] ifFalse:[
		    aGC displayString:line x:x y:ypos
		]
	    ].
	    ypos := ypos + delta
	]
    ]
! !

!DrawText methodsFor:'storing'!

storeInstvarsOn:aStream
    super storeInstvarsOn:aStream.
    aStream nextPutAll:'; setFontFamily:'. fontFamily storeOn:aStream.
    aStream nextPutAll:' face:'. fontFace storeOn:aStream.
    aStream nextPutAll:' style:'. fontStyle storeOn:aStream.
    aStream nextPutAll:' size:'. fontSize storeOn:aStream.
    aStream nextPutAll:'; text:('. text storeOn:aStream.  aStream nextPutAll:')'
! !
