"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#DrawTool
	 instanceVariableNames:'topView drawView menu commandPanel formMenu
				ruler vRuler
				fileBox importBox currentFileName resources buttonGroup'
	 classVariableNames:'ColorList PatternListInMenu PatternList'
	 poolDictionaries:''
	 category:'Applications-DrawTool'
!

DrawTool comment:'

COPYRIGHT (c) 1989-93 by Claus Gittinger
	      All Rights Reserved

an interface to a drawView

$Header: /claus/smalltalk/st/clients/DrawTool/RCS/DrawTool.st,v 1.8 95/02/22 02:50:32 claus Exp $
written summer 89 by claus
'!

!DrawTool class methodsFor:'defaults'!

commandPanelWidth
    ^ 60
!

listOfColorNames
    "return a list of color names to be presented in the fg/bg pulldown menus"

    ^ #('white' 'yellow' 'red' 'magenta' 'blue' 'cyan' 'green' 'black')
!

listOfPatternFileNames
    "return a list of filenames which contain patterns to be
     presented in the pattern/pen pulldown menus"

    ^ #('nopattern.xbm' 'white.xbm'
	'grey94.xbm' 'grey88.xbm' 'grey75.xbm'
	'grey50.xbm' 'grey25.xbm' 'grey12.xbm' 'grey6.xbm'
	'black.xbm'
	'grey75_2.xbm' 'grey50_2.xbm' 'grey25_2.xbm'
	'grey50_3.xbm'
	'hline1.xbm' 'hline2.xbm' 'hline3.xbm'
	'vline1.xbm' 'vline2.xbm' 'vline3.xbm'
	'd1line1.xbm' 'd1line2.xbm' 'd1line3.xbm'
	'd2line1.xbm' 'd2line2.xbm' 'd2line3.xbm'
	'brick1.xbm' 'brick2.xbm'
	'wave1.xbm' 'wave2.xbm'
	'cross1.xbm' 'cross2.xbm')
!

listOfObjectClasses
    "return a collection of classes for which buttons are created in
     the command panel"

    |classes|

    classes := OrderedCollection new.
    classes add:DrawText.
    classes add:DrawLine.
"/    classes add:DrawHLine.
"/    classes add:DrawVLine.
    classes add:DrawRectangle.
    classes add:DrawRoundRectangle.
    classes add:DrawEllipse.
    classes add:DrawPolygon.
    classes add:DrawCurve.

    ^ classes
! !

!DrawTool class methodsFor:'private'!

listOfColors
    "return a list of colors to be presented in the fg/bg pulldown menus.
     Uses names returned by listOfColorNames."

    ColorList isNil ifTrue:[
	ColorList := OrderedCollection new.
	self listOfColorNames do:[:colorName |
	    ColorList add:(Color name:colorName)
	]
    ].
    ^ ColorList
!

listOfPatternsInMenu
    "return a list of patterns to be presented in the pattern/pen pulldown menus.
     Uses names returned by listOfPatternNames."

    |form|

    PatternListInMenu isNil ifTrue:[
	PatternListInMenu := OrderedCollection new.
	PatternList := OrderedCollection new.
	self listOfPatternFileNames do:[:fileName |
	    form := Form fromFile:fileName resolution:100.
	    PatternListInMenu add:form.
	    PatternList add:form
	].
	PatternList at:1 put:nil
    ].
    ^ PatternListInMenu
!

listOfPatterns
    PatternList isNil ifTrue:[
	self listOfPatternsInMenu
    ].
    ^ PatternList
! !

!DrawTool class methodsFor:'startup'!

start
    ^ self open
!

startOn:aFileName
    ^ self openOn:aFileName
!

open
    "startup a DrawTool window with emty contents"

    ^ self new open

    "
     DrawTool open
    "
!

openOn:aFileName
    "startup a DrawTool window, initialize contents from aFileName"

    ^ self new openOn:aFileName

    "
     DrawTool openOn:'aFileName'
    "
! !

!DrawTool methodsFor:'startup'!

open
    |frame panelWidth myname pointerForm protoButton vRulerWidth
     oneMillimeter drawPanel|

    resources := ResourcePack for:self class.

    myname := self class name.

    "
     the outer top view
    "
    topView := StandardSystemView
		label:myname
		icon:(Form fromFile:(myname , '.xbm') resolution:100)
		minExtent:(100 @ 100).

    menu := PullDownMenu in:topView.

    oneMillimeter := Display horizontalPixelPerMillimeter.

    "
     get this form first to see how wide commandPanel should be
    "
    pointerForm := Form fromFile:'Pointer.xbm' resolution:100.
    pointerForm isNil ifTrue:[
	self warn:'DrawTool: cannot find (some of) my bitmaps'.
	panelWidth := 80
    ] ifFalse:[
	protoButton := (RadioButton new) form:pointerForm.
	panelWidth := protoButton width + (oneMillimeter * 2) rounded.
    ].

    "
     the panel containing the buttons
    "
    commandPanel := VerticalPanelView in:topView.
    commandPanel corner:(panelWidth @ 1.0).
    commandPanel layout:#top.
    commandPanel level:0.
    commandPanel borderWidth:0.

    "
     the actual draw view
    "
    drawPanel := View
		     origin:(panelWidth @ (menu height * 2))
		     corner:(1.0 @ 1.0)
		     in:topView.

    "
     horizontal & vertical Rulers
    "
    ruler := Ruler in:drawPanel.

    vRuler := VerticalRuler in:drawPanel "topView".
    vRulerWidth := ruler height.
    vRuler width:vRulerWidth.

    frame := HVScrollableView for:DrawView in:drawPanel "topView".
    frame origin:(0.0 "panelWidth" @ ("menu height * 2 +" ruler height))
	  corner:(1.0 @ 1.0).
    frame rightInset:vRulerWidth+oneMillimeter rounded.

    drawView := frame scrolledView.

    drawView changeFillPattern:nil.
    drawView changeLinePattern:(Form black "fromFile:'black.xbm'").
    drawView changeLineWidth:1.

    ruler origin:[(drawView originRelativeTo:drawPanel "topView") x + drawView borderWidth
		  @ 0 "(menu height * 2)"]
	  corner:(1.0 @ ("menu height * 2 +" ruler height)).
    ruler rightInset:vRulerWidth+oneMillimeter rounded.
    ruler level:-1.

    vRuler origin:(1.0 @ 0.0 "(menu height * 2)")
	   corner:(1.0 @ 1.0).
    vRuler leftInset:vRulerWidth negated - (oneMillimeter // 2).
    vRuler topInset:vRulerWidth + (oneMillimeter // 2) + 1.
    vRuler rightInset:(oneMillimeter // 1 "2").
    vRuler bottomInset:(frame horizontalScrollBar height + oneMillimeter rounded).
    vRuler level:-1.

    "
     now, that we know the dimensions, set the commandPanels origin.
    "
    commandPanel origin:(0 @ ((menu height * 2) + ruler height)).

    "
     want to be notified of scroll operations in the drawView
     (to update the rulers)
    "
    drawView addDependent:self.

    self initMenu.

    "
     the buttons created below are arranged in a group to
     give them one-only behavior
    "
    buttonGroup := RadioButtonGroup new:0.

    buttonGroup add:
	(RadioButton form:pointerForm
		   action:[drawView noCreate]
		       in:commandPanel) turnOn.

    "
     create one button for each drawobject type
    "
    (self class listOfObjectClasses) do:[:elClass |
	buttonGroup add:(RadioButton form:(elClass icon)
				   action:[drawView createObjectWithClass:elClass]
				       in:commandPanel).
    ].

    topView open.
!

openOn:aFileName
    |aStream baseName|

    self open.

    "read file without redraw - done when file has to be read before
     first expose is handled - avoids double redraw"

    baseName := OperatingSystem baseNameOf:aFileName.
    topView label:baseName , ' - ' , (OperatingSystem directoryNameOf:aFileName).
    topView iconName:baseName.

    aStream := FileStream readonlyFileNamed:aFileName.
    aStream notNil ifTrue:[
	drawView withoutRedrawFileInContentsFrom:aStream.
	aStream close.
	currentFileName := aFileName
    ]
! !

!DrawTool methodsFor:'menu definition'!

initMenu
    |labels|

    labels := resources array:#(
			       'file'
			       'edit'
			       'font' 
			       'form'
			       'extra'
			       'pattern'
			       'line'
			       'pen' 
			       'foreground'
			       'background').
    menu labels:labels.
    menu selectors:#(file edit font form extra pattern line pen fg bg).

    menu at:#file 
	    putLabels:(resources array:#(
					'new' 
					'open ..' 
					'load ..' 
					'-' 
					'import ..' 
					'-' 
					'save' 
					'save as ...' 
					'-' 
					'print'
					'-' 
					'quit'))
	    selectors:#(doNew doOpen doLoad nil doImport nil 
			doSave doSaveAs nil doPrint
			nil finish)
	     receiver:self.

    menu at:#edit 
	    putLabels:(resources array:#(
					'copy' 
					'cut' 
					'paste'
"
					'-'
					'edit bitmap'
"
					))
	    selectors:#(copySelection 
			deleteSelection 
			pasteBuffer 
"
			nil 
			editBitmap
"
		       )
	     receiver:drawView.

    menu at:#font 
	    putLabels:(resources array:#(
					'larger' 
					'smaller' 
					'-'
					'normal' 
					'bold' 
					'italic' 
					'bold italic' 
					'-'
					'font panel'))
	    selectors:#(largerFont
			smallerFont
			nil
			normalFont
			boldFont
			italicFont
			boldItalicFont
			nil
			showFontPanel)
	     receiver:drawView.

    (Smalltalk language == #english) ifTrue:[
	menu at:#form 
		putLabels:(resources array:#(
					    'grid on' 
					    'show grid'
					    '-'
					    'millimeter'
"/                                            'hide ruler'
					    '-'
					    'page size'
					    'zoom'
					    ))
		selectors:#(alignOn 
			    showGrid 
			    nil 
			    millimeterMetric 
"/                            scaleOff 
			    nil 
			    pageSize
			    zoom)
		 receiver:self
    ] ifFalse:[
	menu at:#form
		putLabels:(resources array:#(
					    'grid on' 
					    'show grid'
					    '-'
					    'inches'
"/                                            'hide ruler'
					    '-'
					    'page size'
					    'zoom'
					    ))
		selectors:#(alignOn 
			    showGrid 
			    nil 
			    inchMetric 
"/                            scaleOff 
			    nil 
			    pageSize
			    zoom)
		 receiver:self
    ].
    formMenu := menu menuAt:#form.

    formMenu subMenuAt:#pageSize
		   put:(PopUpMenu
			   labels:(resources array:#(
						     'a6'       
						     'a5'       
						     'a4'       
						     'a3'       
						     'a2'
						     'a1'
						     'letter'
						     '-'
						     'a6 landscape'       
						     'a5 landscape'       
						     'a4 landscape'       
						     'a3 landscape'       
						     'a2 landscape'
						     'a1 landscape'
						     'letter landscape'
						    ))
			   selectors:#documentFormat:
			   args:#('a6' 'a5' 'a4' 'a3' 'a2' 'a1' 'letter' nil
				  'a6l' 'a5l' 'a4l' 'a3l' 'a2l' 'a1l' 'letterl' )
			   receiver:self
				for:formMenu 
			).

    formMenu subMenuAt:#zoom 
		   put:(PopUpMenu
			   labels:(resources array:#(
						     '100 %'       
						     '-'       
						     '25 %'       
						     '50 %'       
						     '75 %'
						     '-'
						     '150 %'
						     '200 %'
						     '-'
						     'other ...'       
						    ))
			   selectors:#setZoom:
			   args:#(100 nil 25 50 75 nil 150 200 nil nil)
			   receiver:self
				for:formMenu 
			).

    menu at:#extra 
	    putLabels:(resources array:#(
					'to front' 
					'to back'
					'-' 
					'flip horizontal' 
					'flip vertical' 
					'rotate cw' 
					'rotate ccw' 
					'-' 
					'align left' 
					'align right' 
					'align top' 
					'align bottom' 
					'-' 
					'select all' 
					'-' 
					'group' 
					'ungroup'))
	    selectors:#(selectionToFront selectionToBack
			nil flipHorizontal flipVertical
			    rotate90 rotate270
			nil selectionAlignLeft selectionAlignRight
			    selectionAlignTop selectionAlignBottom
			nil selectAll
			nil group ungroup)
	     receiver:drawView.

    menu at:#pattern 
	    putMenu:(PatternMenu
			patterns:(self class listOfPatternsInMenu)
			selectors:#changeFillPattern:
			args:(self class listOfPatterns)
			receiver:drawView
			for:menu).

    menu at:#line 
"
	    putLabels:#('1' 
			'2' 
			'3' 
			'4' 
			'5' 
			'6' 
			'7' 
			'8' 
			'9' 
			'10' 
			'15' 
			'20')
	    selectors:#changeLineWidth:
		 args:#(1 2 3 4 5 6 7 8 9 10 15 20)
	     receiver:drawView.
"
	    putMenu:(LineMenu
			lineWidths:#(1 2 3 4 5 6 7 8 9 10 15 20)
			selectors:#changeLineWidth:
			args:#(1 2 3 4 5 6 7 8 9 10 15 20)
			receiver:drawView
			for:menu).

    menu at:#pen 
	    putMenu:(PatternMenu
			patterns:(self class listOfPatternsInMenu)
			selectors:#changeLinePattern:
			args:(self class listOfPatterns)
			receiver:drawView
			for:menu).

    menu at:#fg 
	    putMenu:(PatternMenu
			patterns:(self class listOfColors)
			selectors:#changeForeground:
			args:(self class listOfColors)
			receiver:drawView
			for:menu).

    menu at:#bg 
	    putMenu:(PatternMenu
			patterns:(self class listOfColors)
			selectors:#changeBackground:
			args:(self class listOfColors)
			receiver:drawView
			for:menu).
! !

!DrawTool methodsFor:'changes'!

update:something with:someArgument from:changedObject
    "since we installed ourself as dependent of the drawView,
     we will be notified here about scroll actions. Pass scroll activity
     to rulers."

    changedObject == drawView ifTrue:[
	something == #originOfContents ifTrue:[
	    ruler viewOrigin:(drawView viewOrigin x @ 0).
	    vRuler viewOrigin:(0 @ drawView viewOrigin y).
	]
    ]
! !

!DrawTool methodsFor:'user interaction'!

finish
    fileBox notNil ifTrue:[fileBox destroy. fileBox := nil].
    importBox notNil ifTrue:[importBox destroy. importBox := nil].
    drawView := menu := commandPanel := nil.
    formMenu := ruler := resources := buttonGroup := nil.
    topView destroy. 
    "
     notReached
    "
    topView := nil.
!

doNew
    "clear"

    drawView removeAll
!

openFile:aFileName new:new
    |aStream baseName|

    aStream := FileStream readonlyFileNamed:aFileName.
    aStream notNil ifTrue:[
	drawView fileInContentsFrom:aStream redraw:true new:new.
	aStream close.
	currentFileName := aFileName.
	baseName := OperatingSystem baseNameOf:aFileName.
	topView label:baseName , ' - ' , (OperatingSystem directoryNameOf:aFileName).
	topView iconName:baseName
    ]
!

doOpen
    "clear and load a new drawing"

    fileBox isNil ifTrue:[
	fileBox := FileSelectionBox
			title:''
			okText:''
			abortText:(resources string:'abort')
			action:[nil].
    ].
    fileBox pattern:'*.draw'.
    fileBox title:(resources string:'open file:').
    fileBox action:[:fileName | self openFile:fileName new:true].
    fileBox okText:(resources string:'open').

    currentFileName notNil ifTrue:[
	(currentFileName startsWith:'/') ifTrue:[
	    fileBox directory:(OperatingSystem directoryNameOf:currentFileName)
	]
    ].
    fileBox showAtPointer
!

doLoad
    "load a drawing and merge into the current one"

    fileBox isNil ifTrue:[
	fileBox := FileSelectionBox
			title:''
			okText:''
			abortText:(resources string:'abort')
			action:[nil].
    ].
    fileBox pattern:'*.draw'.
    fileBox title:(resources string:'load objects from file:').
    fileBox action:[:fileName | self openFile:fileName new:false].
    fileBox okText:(resources string:'load').

    currentFileName notNil ifTrue:[
	(currentFileName startsWith:'/') ifTrue:[
	    fileBox directory:(OperatingSystem directoryNameOf:currentFileName)
	]
    ].
    fileBox showAtPointer
!

doClose
    ^ self
!

saveAs:aFileName
    "common helper for saving"

    |aStream|

    aStream := FileStream newFileNamed:aFileName.
    aStream notNil ifTrue:[
	drawView storeContentsOn:aStream.
	aStream close
    ].
    currentFileName := aFileName
!

doSave
    "save in my own private format - in current file"

    currentFileName notNil ifTrue:[
	self saveAs:currentFileName
    ] ifFalse:[
	self doSaveAs
    ]
!

doSaveAs
    "save in my own private format - let user specify a filename"

    fileBox isNil ifTrue:[
	fileBox := FileSelectionBox
			title:''
			okText:''
			abortText:(resources string:'abort')
			action:[nil].
    ].
    fileBox pattern:'*.draw'.
    fileBox title:(resources string:'save as:').
    fileBox action:[:fileName | self saveAs:fileName].
    fileBox okText:(resources string:'save').
    fileBox showAtPointer
!

import:aFileName
    "bitmap/image import"

    |img|

    drawView windowGroup withCursor:Cursor wait do:[
	Processor activeProcess withPriority:(Processor userBackgroundPriority)
	do:[
	    ObjectMemory allocationFailureSignal handle:[:ex |
		 self warn:'no memory for image'.
		 ex return
	    ] do:[
		img := Image fromFile:aFileName.
		img notNil ifTrue:[
		    drawView addSelected:(DrawImage new image:img).
		] ifFalse:[
		    aFileName asFilename isReadable ifTrue:[
			 self warn:('unknown image format: ', aFileName)
		   ] ifFalse:[
			self warn:(aFileName , ' unreadable')
		   ]
		]
	    ].
	].
    ].
!

doImport
    importBox isNil ifTrue:[
	importBox := FileSelectionBox
			title:''
		       okText:''
		    abortText:(resources string:'abort')
		       action:[nil].
    ].
    importBox pattern:'*'.
    importBox title:(resources string:'Filename of image:').
    importBox action:[:fileName | self import:fileName].
    importBox okText:(resources string:'import').
    importBox showAtPointer
!

doPrint
    |gc printStream|

    (Smalltalk includesKey:#PSGraphicsContext) ifFalse:[
	^ self warn:'not implemented (need PSGraphicsContext)'
    ].
    PSGraphicsContext isNil ifTrue:[
	^ self warn:'not implemented (need PSGraphicsContext)'
    ].
    (Printer == PostscriptPrinterStream) ifFalse:[
	^ self warn:'currently only postscript printing supported'
    ].

    printStream := PipeStream writingTo:(Printer printCommand).

    gc := PSGraphicsContext on:printStream extent:(drawView widthOfContents @ drawView heightOfContents).
    [
	drawView redrawObjectsOn:gc
    ] valueNowOrOnUnwindDo:[gc close].
!

documentFormat:aFormat
    drawView documentFormat:aFormat.
    ruler paperWidthMM:(drawView widthOfContentsInMM).
    vRuler paperHeightMM:(drawView heightOfContentsInMM)
!

setZoom:zoom 
    |zoomFactor|

    zoomFactor := zoom.
    zoom isNil ifTrue:[
	|box zoomString|

	box := EnterBox title:'zoom factor (percent):'.
	box okText:'change'.
	zoomString := box request.
	zoomString isNil ifTrue:[^ self].
	zoomFactor := Number readFromString:zoomString onError:[100].
    ].

    drawView zoom:zoomFactor / 100.
    ruler scale:zoomFactor / 100.
    vRuler scale:zoomFactor / 100.
!

scaleOn
    formMenu labelAt:#scaleOn put:(resources string:'hide ruler').
    formMenu selectorAt:#scaleOn put:#scaleOff.
    drawView showScale
!

scaleOff
    formMenu labelAt:#scaleOff put:(resources string:'show ruler').
    formMenu selectorAt:#scaleOff put:#scaleOn.
    drawView hideScale
!

millimeterMetric
    formMenu labelAt:#millimeterMetric put:(resources name:'inches').
    formMenu selectorAt:#millimeterMetric put:#inchMetric.
    ruler metric:#mm.
    vRuler metric:#mm.
    drawView millimeterMetric.
    ^ self
!

inchMetric    
    formMenu labelAt:#inchMetric put:(resources string:'millimeter').
    formMenu selectorAt:#inchMetric put:#millimeterMetric.
    ruler metric:#inch.
    vRuler metric:#inch.
    drawView inchMetric.
    ^ self
!

alignOn
    formMenu labelAt:#alignOn put:(resources string:'grid off').
    formMenu selectorAt:#alignOn put:#alignOff.
    drawView alignOn
!

alignOff
    formMenu labelAt:#alignOff put:(resources string:'grid on').
    formMenu selectorAt:#alignOff put:#alignOn.
    drawView alignOff
!

showGrid
    formMenu labelAt:#showGrid put:(resources string:'hide grid').
    formMenu selectorAt:#showGrid put:#hideGrid.
    drawView showGrid.
!

hideGrid
    formMenu labelAt:#hideGrid put:(resources string:'show grid').
    formMenu selectorAt:#hideGrid put:#showGrid.
    drawView hideGrid.
! !
