'From Smalltalk/X, Version:2.10.2 on 23-may-1994 at 13:22:46'!

GLObjectDemoView subclass:#GLCubeDemoView2
	 instanceVariableNames:'rotLightMatrix lightPos light model material normals'
	 classVariableNames:''
	 poolDictionaries:''
	 category:'Demos-3D'
!

!GLCubeDemoView2 class methodsFor:'startup'!

open 
    "check if display has rgb features"

    Display supportsRGB ifFalse:[
	^ self warn:'sorry, but this demo requires a GL with RGB support'
    ].
    Display supportsLight ifFalse:[
	^ self warn:'sorry, but this demo requires a GL with light support'
    ].
    ^ super open
! !

!GLCubeDemoView2 class methodsFor:'defaults'!

defaultExtent
    ^ 500 @ 500
!

defaultLabel
    "used by superclass"

    ^ 'GL solid cube & light'
! !

!GLCubeDemoView2 methodsFor:'drawing'!

redraw
    realized ifFalse:[^ self].

    "
     do not draw, if iconified
    "
    shown ifTrue:[
	self perspectiveFovy:450 aspect:1.0 near:1.0 far:200.0.

	self mMode:#viewing.

	self lmbind:#material   index:1.
	self lmbind:#light0     index:1.
	self lmbind:#lightModel index:1.

	self colorRed:0 green:0 blue:0.
	self clear.

	"rotate the light"

"
	self pushMatrix.
	self loadMatrix:IdentityMatrix.
	self rotateI:1 axis:#x.
	self rotateI:2 axis:#y.
	self multMatrix:rotLightMatrix.
	self lmdef:#light index:1 np:0 props:lightPos.
	self getMatrix:rotLightMatrix.
	self popMatrix.
 "

	"
	 rotate & draw the object
	"
	self pushMatrix.
	self translateZ:trans.
	self rotateX:angleX y:angleY z:angleZ.
	self callObject:objectId.
	self popMatrix.

	self swapBuffers.
    ].

    "
     on systems without processes, install a timed-block
     for the next frame.
    "
    redrawBlock notNil ifTrue:[
	Processor addTimedBlock:redrawBlock afterSeconds:deltaT
    ]
!

drawObject
    self lmbind:#material   index:1.

    self backFace:true.

    "front side"
    self beginPolygon.  
    self n3f:(normals at:1).
    self v3f:(points at:1).
    self v3f:(points at:2).
    self v3f:(points at:3).
    self v3f:(points at:4).
    self endPolygon. 

    "back side"
    self beginPolygon. 
    self n3f:(normals at:6).
    self v3f:(points at:6).
    self v3f:(points at:5).
    self v3f:(points at:8).
    self v3f:(points at:7).
    self endPolygon. 

    "right side"
    self beginPolygon. 
    self n3f:(normals at:2).
    self v3f:(points at:2).
    self v3f:(points at:6).
    self v3f:(points at:7).
    self v3f:(points at:3).
    self endPolygon. 

    "left side"
    self beginPolygon. 
    self n3f:(normals at:3).
    self v3f:(points at:1).
    self v3f:(points at:4).
    self v3f:(points at:8).
    self v3f:(points at:5).
    self endPolygon. 

    "top side"
    self beginPolygon. 
    self n3f:(normals at:5).
    self v3f:(points at:3).
    self v3f:(points at:7).
    self v3f:(points at:8).
    self v3f:(points at:4).
    self endPolygon. 

    "bottom side"
    self beginPolygon. 
    self n3f:(normals at:4).
    self v3f:(points at:1).
    self v3f:(points at:5).
    self v3f:(points at:6).
    self v3f:(points at:2).
    self endPolygon. 
!

defineLightAndMaterial
    self lmdef:#light      index:1 np:0 props:light.
    self lmdef:#material   index:1 np:0 props:material.
    self lmdef:#lightModel index:1 np:0 props:model.
! !

!GLCubeDemoView2 methodsFor:'initialization'!

setupPoints
    points := Array new:8.
    points at:1 put:(Point3D x:-1.0 y:-1.0 z: 1.0).   "front bot-left"
    points at:2 put:(Point3D x: 1.0 y:-1.0 z: 1.0).   "      bot-right"
    points at:3 put:(Point3D x: 1.0 y: 1.0 z: 1.0).   "      top-right"
    points at:4 put:(Point3D x:-1.0 y: 1.0 z: 1.0).   "      top-left"
    points at:5 put:(Point3D x:-1.0 y:-1.0 z:-1.0).   "rear"
    points at:6 put:(Point3D x: 1.0 y:-1.0 z:-1.0).   
    points at:7 put:(Point3D x: 1.0 y: 1.0 z:-1.0).   
    points at:8 put:(Point3D x:-1.0 y: 1.0 z:-1.0).  

    normals := Array new:6.
    normals at:1 put:(Point3D x: 0.0 y: 0.0 z:-1.0).   "front"
    normals at:2 put:(Point3D x:-1.0 y: 0.0 z: 0.0).   "right"
    normals at:3 put:(Point3D x: 1.0 y: 0.0 z: 0.0).   "left"
    normals at:4 put:(Point3D x: 0.0 y: 1.0 z: 0.0).   "bottom"
    normals at:5 put:(Point3D x: 0.0 y:-1.0 z: 0.0).   "top"
    normals at:6 put:(Point3D x: 0.0 y: 0.0 z: 1.0).   "rear"
!

realize
    super realize.
    self defineLightAndMaterial
!

recreate
    super recreate.
    self defineLightAndMaterial
!

initialize
    super initialize.
    type := #rgbDoubleBuffer.

    deltaRot := 0.75.
    deltaTrans := 0.05.
    transZ := false.

    rotLightMatrix := IdentityMatrix copy.

    lightPos := FloatArray withAll:#( 102.0 "POSITION"     0.0 0.0 1.0 0.0 
					0.0 "END" ).

    light :=    FloatArray withAll:#(   2.0 "AMBIENT"      0.1 0.1 0.1
				      101.0 "LCOLOR"       0.5 1.0 1.0
				      102.0 "POSITION"     90.0 90.0 -150.0 0.0
					0.0 "END" ).


    model :=    FloatArray withAll:#(   2.0 "AMBIENT"      1.0 1.0 1.0
				      201.0 "LOCALVIEWER"  0.0
					0.0 "END" ).

    material := FloatArray withAll:#(   5.0 "SHININESS"    30.0
					4.0 "SPECULAR"     0.2 0.2 0.2
					0.0 "END" ).
! !
