"
 COPYRIGHT (c) 1994 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

GLXView subclass:#GLObjectDemoView
	 instanceVariableNames:'trans angleX angleY angleZ deltaTrans deltaRot minTrans maxTrans
		points rotX rotY rotZ transZ objectId
		redrawBlock myProcess deltaT fullSpeed'
	 classVariableNames:''
	 poolDictionaries:''
	 category:'Demos-3D'
!

GLObjectDemoView comment:'
COPYRIGHT (c) 1994 by Claus Gittinger
	      All Rights Reserved
'!

!GLObjectDemoView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1994 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /claus/smalltalk/st/clients/GLdemos/RCS/GLObjectDemo.st,v 1.7 95/03/09 03:23:18 claus Exp $
"
!

documentation
"
    abstract class for rotating objects animation demos.
    The object is drawn once (during init) and thereby defined
    as a GL object. Then, in regular intervals, the object is
    simply called for (i.e. drawing is done completely in GL).

    Concrete subclasses should redefine #drawObject and #setupPoints.

    start with:
	GLWireCubeDemoView open
	GLCubeDemoView open
	GLTetraDemoView open

    Notice: for ST/X configurations without multiple processes
	    (i.e. pureEventDriven systems) this demo installs
	    a timeout-block to do the animation.

	    For normal systems, the animation is done by a
	    low-priority background process, sending #animate
	    in regular intervals.
"
! !

!GLObjectDemoView class methodsFor:'defaults'!

defaultExtent
    ^ 300 @ 300
! !

!GLObjectDemoView methodsFor:'drawing'!

animate
    "rotate and/or translate; then redraw"

    rotX ifTrue:[angleX := angleX + deltaRot].
    rotY ifTrue:[angleY := angleY + deltaRot].
    rotZ ifTrue:[angleZ := angleZ + deltaRot].
    transZ ifTrue:[
	trans := trans + deltaTrans.
	((trans > maxTrans) or:[trans < minTrans]) ifTrue:[
	    deltaTrans := deltaTrans negated
	].
    ].
    self redraw.
!

drawFrame
    "
     let GL draw the previously defined object
    "
    self callObject:objectId.
!

redraw 
    realized ifTrue:[

	"
	 do not draw, if iconified
	"
	shown ifTrue:[
	    "
	     clear the drawing area
	    "
	    (type == #colorIndexDoubleBuffer
	    or:[type == #colorIndex]) ifTrue:[
		self color:Black
	    ] ifFalse:[
		self colorRed:0 green:0 blue:0.
	    ].
	    self clear.

	    "
	     save current rotation/translation
	    "
	    self pushMatrix.

	    "
	     rotate/translate
	    "
	    self translateZ:trans.
	    self rotateX:angleX y:angleY z:angleZ.

	    self drawFrame.

	    "
	     restore rotation/translation
	    "
	    self popMatrix.
	    "
	     swap buffers for next round
	    "
	    self swapBuffers.
	].
	"
	 on systems without processes, install a timed-block
	 for the next frame.
	"
	redrawBlock notNil ifTrue:[
	    Processor addTimedBlock:redrawBlock afterSeconds:deltaT
	]
    ]
!

setupPoints
    "this should be redefined in concrete subclasses,
     to setup the points of the animated object."

    self subclassResponsibility
!

drawObject
    "this should be redefined in concrete subclasses,
     to draw the object (using GL functions only)."

    self subclassResponsibility
! !

!GLObjectDemoView methodsFor:'initialization'!

unrealize 
    redrawBlock notNil ifTrue:[
	Processor removeTimedBlock:redrawBlock.
    ] ifFalse:[
	myProcess notNil ifTrue:[
	    myProcess terminate.
	    myProcess := nil
	]
    ].
    super unrealize.
    KnownObjects remove:objectId
!

initialize
    super initialize.
    type := #colorIndexDoubleBuffer.

    ProcessorScheduler isPureEventDriven ifTrue:[
	redrawBlock := [self animate].
    ].

    "
     time delta between frames
    "
    fullSpeed := false.
    deltaT := 1/60.

    "
     initial rotation and translation
    "
    angleX := angleY := angleZ := 0.0.
    trans := -5.0.
    "
     delta-rotation/translation for each frame
    "
    deltaRot := 3.0.
    deltaTrans := 0.2.
    "
     bounds for translation
    "
    minTrans := -20.0.
    maxTrans := -3.0.

    "
     yes, we rotate and translate
    "
    rotX := true.
    rotY := true.
    rotZ := true.
    transZ := true.

    self setupPoints
!

initializeMiddleButtonMenu
    self middleButtonMenu:(PopUpMenu labels:#('\c rotate x'
					      '\c rotate y'
					      '\c rotate z'
					      '-'
					      '\c translate Z'
					      '-'
					      '\c stop'
					      '-'
					      'faster'
					      'slower'
"/                                              '\c full speed'
					     )
				  selectors:#(rotX: 
					      rotY: 
					      rotZ: 
					      nil 
					      transZ: 
					      nil 
					      stay: 
					      nil 
					      faster 
					      slower
"/                                              fullSpeed:
					     )
				   receiver:self
					for:self).
    middleButtonMenu checkToggleAt:#rotX: put:rotX.
    middleButtonMenu checkToggleAt:#rotY: put:rotY.
    middleButtonMenu checkToggleAt:#rotZ: put:rotZ.
    middleButtonMenu checkToggleAt:#transZ: put:transZ.
    middleButtonMenu checkToggleAt:#stay: put:false.
    middleButtonMenu checkToggleAt:#fullSpeed: put:fullSpeed.
!

setupViewing
    "
     define perspective (see GL manual)
    "
    self perspectiveFovy:450 aspect:1.0 near:1.0 far:200.0. 
!

realize 
    super realize .

    self setupViewing. 

    "
     define a new GL object
    "
    objectId := KnownObjects size + 1.
    KnownObjects add:objectId.
    self makeObject:objectId.
    self drawObject.
    self closeObject.

    "
     on pure event driven systems (redrawBlock notNil)
     install it via the timeout mechanism.
     On normal systems, start a low-prio thread to do
     the animation.
    "
    redrawBlock notNil ifTrue:[
	Processor addTimedBlock:redrawBlock afterSeconds:deltaT
    ] ifFalse:[
	myProcess := [
	    |d|

	    [true] whileTrue:[
		shown ifTrue:[
		    fullSpeed ifFalse:[
			(Delay forSeconds:deltaT) wait.
		    ].
		    self animate
		] ifFalse:[
		    "
		     delay a bit longer if iconified
		    "
		    (Delay forSeconds:deltaT * 4) wait.
		]
	    ]
	] forkAt:4.
	myProcess name:'GL animation'
    ]    
!

reinitialize
    super reinitialize.
    self realize. 
! !

!GLObjectDemoView methodsFor:'menu actions'!

rotX:aBoolean
    "turn on/off x-rotation"

    rotX := aBoolean.
    middleButtonMenu checkToggleAt:#stay: put:(rotX or:[rotY or:[rotZ or:[transZ]]]) not.
!

rotY:aBoolean
    "turn on/off y-rotation"

    rotY := aBoolean.
    middleButtonMenu checkToggleAt:#stay: put:(rotX or:[rotY or:[rotZ or:[transZ]]]) not.
!

rotZ:aBoolean
    "turn on/off z-rotation"

    rotZ := aBoolean.
    middleButtonMenu checkToggleAt:#stay: put:(rotX or:[rotY or:[rotZ or:[transZ]]]) not.
!

transZ:aBoolean
    "turn on/off z-translation"

    transZ := aBoolean.
    middleButtonMenu checkToggleAt:#stay: put:(rotX or:[rotY or:[rotZ or:[transZ]]]) not.
!

stay:aBoolean
    "stop - turn on/off all rotations/translations"

    rotX := rotY := rotZ := transZ := aBoolean not.
    middleButtonMenu checkToggleAt:#rotX: put:rotX.
    middleButtonMenu checkToggleAt:#rotY: put:rotY.
    middleButtonMenu checkToggleAt:#rotZ: put:rotZ.
    middleButtonMenu checkToggleAt:#transZ: put:transZ.
!

fullSpeed:aBoolean
    "turn on/off delays"

    fullSpeed := aBoolean.
!

faster
    "faster - make delta-rotation a bit bigger"

    deltaRot := deltaRot * 2.
    deltaTrans := deltaTrans * 2
!

slower 
    "slower - make delta-rotation a bit smaller"

    deltaRot := deltaRot / 2.
    deltaTrans := deltaTrans / 2
! !
