'From Smalltalk/X, Version:2.6.5 on 13-Jun-1993 at 8:50:47'!

View subclass:#GLXView
	 instanceVariableNames:'type'
	 classVariableNames:'IdentityMatrix
			     Black Red Green Yellow Blue Magenta Cyan White
			     KnownObjects'
	 poolDictionaries:''
	 category:'Views-3D'
!

!GLXView class methodsFor:'documentation'!

documentation
"
    This view (and therefore all subclasses) support GLX type of
    3D rendering. This is a prototype implementation with just a
    few things implemented; to show how things could be and what is
    possible.

    GLXViews on non-SGI workstations will only support a subset
    of the GL functionality (i.e. no lighting, no Z-buffer, no textures
    etc.). The GL simulation used on those machines is based on the
    vogl PD library.

    To understand the individual methods, please refer to the GL programmers
    manual or the OpenGL documentation. Documentation for some functions is
    also found in the vogl documentation.

    Notice, not all possible functions (see GLXWorkstation) have been 
    implemented here - new ones will be added as required.
    (Actually, for every GL-method in GLXWorkstation, a corresponding
     method is required here.)

    written jun 93 by claus;
    many, many additional methods (for almost full set of graphic calls)
    dec 93 by jeff
"
! !

!GLXView class methodsFor:'initialization'!

initialize
    super initialize.
    IdentityMatrix := FloatArray withAll:#(  1.0 0.0 0.0 0.0
					     0.0 1.0 0.0 0.0
					     0.0 0.0 1.0 0.0 
					     0.0 0.0 0.0 1.0 ).

    "
     common color indices as class variables
     Notice: these could (in theory) be redefined using mapColor.
     however, it is recommended to leave those 8 colors unchanged.
     start private colors with index above 7
    "
    Black := 0.
    Red := 1.
    Green := 2.
    Yellow := 3.
    Blue := 4.
    Magenta := 5.
    Cyan := 6.
    White := 7.

    KnownObjects isNil ifTrue:[
	KnownObjects := OrderedCollection new.
    ]
! !

!GLXView class methodsFor:'start'!

defaultLabel
    ^ 'a GLXView'
!

defaultIcon
    ^ Image fromFile:'bitmaps/GLX.xbm'
!

defaultMinExtent
    ^ self defaultExtent
!

defaultMaxExtent
    ^ self defaultExtent
!

open
    "default open for GLX views: create a standardSystemView around,
     and open that one"

    |top v|

    (Display isKindOf:GLXWorkstation) ifFalse:[
	self warn:'sorry, but this view requires a display capable of GL drawing'.
	^ nil
    ].
    top := StandardSystemView new.
    top extent:(self defaultExtent).
    top minExtent:(self defaultMinExtent).
    top maxExtent:(self defaultMaxExtent).
    top label:(self defaultLabel).
    top icon:(self defaultIcon).
    v := self in:top.
    v origin:0@0 extent:1.0@1.0.
    top open.
    ^ v
! !

!GLXView class methodsFor:'constants'!

defaultMaterial
    ^ FloatArray withAll:#(
			   7.0 "ALPHA"      1.0             "/ 0.0 .. 1.0
			   2.0 "AMBIENT"    0.2  0.2  0.2   "/ 0.0 .. 1.0
			   3.0 "DIFFUSE"    0.8  0.8  0.8   "/ 0.0 .. 1.0
			   1.0 "EMISSION"   0.0  0.0  0.0   "/ 0.0 .. 1.0
			   4.0 "SPECULAR"   0.0  0.0  0.0   "/ 0.0 .. 1.0
			   5.0 "SHININESS"  0.0             "/ 0.0 .. 128.0
			   0.0 "END"
			  )
!

identityMatrix
    ^ IdentityMatrix
!

white
    "return the index of the white color - this has a predefined
     index in the colorMap. Only usefull for colorIndex modes."

    ^ White
!

yellow
    "return the index of the yellow color - this has a predefined
     index in the colorMap. Only usefull for colorIndex modes."

    ^ Yellow
!

red
    "return the index of the red color - this has a predefined
     index in the colorMap. Only usefull for colorIndex modes."

    ^ Red
!

blue
    "return the index of the blue color - this has a predefined
     index in the colorMap. Only usefull for colorIndex modes."

    ^ Blue
!

green
    "return the index of the green color - this has a predefined
     index in the colorMap. Only usefull for colorIndex modes."

    ^ Green
!

cyan 
    "return the index of the cyane color - this has a predefined
     index in the colorMap. Only usefull for colorIndex modes."

    ^ Cyan
!

magenta 
    "return the index of the magenta color - this has a predefined
     index in the colorMap. Only usefull for colorIndex modes."

    ^ Magenta
!

black
    "return the index of the black color - this has a predefined
     index in the colorMap. Only usefull for colorIndex modes."

    ^ Black
! !

!GLXView methodsFor:'initialization'!

initialize
    super initialize.
    type := #colorIndexSingleBuffer
! !

!GLXView methodsFor:'realization'!

physicalCreate
    "redefined create, to make a GLX window instead of a normal
     one (should extract the createWindow from View and just
     redefine that on to get rid of all the other stuff here)"

    drawableId := device createGLXWindowFor:self 
				       left:left top:top 
				      width:width height:height
				       type:type.

    extentChanged := false.
    originChanged := false.

    borderShape notNil ifTrue:[
	device setWindowBorderShape:(borderShape id) in:drawableId
    ].
    viewShape notNil ifTrue:[
	device setWindowShape:(viewShape id) in:drawableId
    ].
    backed notNil ifTrue:[
	device setBackingStore:backed in:drawableId
    ].
    saveUnder ifTrue:[
	device setSaveUnder:true in:drawableId
    ].
!

destroy
    "view is about to be destroyed - unlink from GLX first"

    drawableId notNil ifTrue:[
	device unlinkGLXView:drawableId.
    ].
    super destroy
!

realize
    superView isNil ifTrue:[
	"
	 we have to trap this situation, since GL will
	 close down the X-connection if a GL-topview
	 is closed (i.e. GLXView new open).
	"
	self error:'GLXViews need a superview'.
	^ self
    ].
    super realize.
    self enableEvent:#expose.
    "
     SGI-GL does not generate mapped/unmapped etc. events.
     (which is bad)
    "
    shown := true
! !

!GLXView methodsFor:'events '!

configureX:x y:y width:w height:h
    "view was resized - have to tell GL about it"

    super configureX:x y:y width:w height:h.
    device glxReshapeViewPortIn:drawableId.
! !

!GLXView methodsFor:'low level data transfer'!

n3f:arrayOf3FloatsWithNormal
    "load vertex normal"

    ^ device glxN3f:arrayOf3FloatsWithNormal in:drawableId
!

c3f:arrayOf3FloatsWithColor
    "load color (r, g, b) vector"

    ^ device glxC3f:arrayOf3FloatsWithColor in:drawableId
!

v3f:arrayOf3Floats
    "load vertex from an array"

    ^ device glxV3f:arrayOf3Floats in:drawableId
!

v3fX:x y:y z:z
    "load vertex from individual x, y and z coordinates.
     If the x/y/z values are computed, this avoids the creation
     of a temporary array."

    ^ device glxV3fX:x y:y z:z in:drawableId
!

t2f:arrayOf2Floats
    "load texture coordinate from an array"

    ^ device glxT2f:arrayOf2Floats in:drawableId
!

v2f:arrayOf2Floats
    "load vertex from an array; z is taken as 0.0"

    ^ device glxV2f:arrayOf2Floats in:drawableId
!

v2fX:x y:y 
    "load vertex from individual x, y coordinates; z is taken as 0.0.
     If the x/y values are computed, this avoids the creation
     of a temporary array."

    ^ device glxV2fX:x y:y in:drawableId
!

vOrigin
    "load 0/0/0 vertex.
     This is the same as v3f:#(0 0 0), but sligtly faster.
     Provided since its a very common vertex."

    ^ device glxVOriginIn:drawableId
!

vUnitX
    "load 1/0/0 vertex.
     This is the same as v3f:#(1 0 0), but sligtly faster.
     Provided since its a very common vertex."

    ^ device glxVUnitXIn:drawableId
!

vUnitY
    "load 0/1/0 vertex.
     This is the same as v3f:#(0 1 0), but sligtly faster.
     Provided since its a very common vertex."

    ^ device glxVUnitYIn:drawableId
!

vUnitZ
    "load 0/0/1 vertex.
     This is the same as v3f:#(0 0 1), but sligtly faster.
     Provided since its a very common vertex."

    ^ device glxVUnitZIn:drawableId
! !

!GLXView methodsFor:'low level objects'!

makeObject:id
    "start an object definition"

    ^ device glxMakeObject:id in:drawableId
!

closeObject
    "end an object definition"

    ^ device glxCloseObjectIn:drawableId
!

deleteObject:id
    "delete an object definition"

    ^ device glxDelecteObject:id in:drawableId
!

callObject:id
    "perform commands from an object"

    ^ device glxCallObject:id in:drawableId
! !

!GLXView methodsFor:'low level viewing setup'!

perspectiveFovy:fovy aspect:aspect near:near far:far
    "define perspective projection"

    ^ device glxPerspectiveFovy:fovy aspect:aspect near:near far:far in:drawableId
!

lookatVx:vx vy:vy vz:vz px:px py:py pz:pz twist:twist
    "define viewing transformation"

    ^ device glxLookatVx:vx vy:vy vz:vz 
		      px:px py:py pz:pz 
		   twist:twist
		      in:drawableId
!

windowLeft:left right:right top:top bottom:bottom near:near far:far 
    "define viewing transformation"

    ^ device glxWindowLeft:left right:right 
		       top:top bottom:bottom 
		      near:near far:far 
			in:drawableId
! !

!GLXView methodsFor:'low level matrix operations'!

loadIdentityMatrix
    "load identity matrix into top of matrix stack"

    ^ device glxLoadMatrix:IdentityMatrix in:drawableId
!

loadMatrix:arrayOf16Floats
    "load transformation matrix"

    ^ device glxLoadMatrix:arrayOf16Floats in:drawableId
!

multMatrix:arrayOf16Floats
    "multiplication matrix"

    ^ device glxMultMatrix:arrayOf16Floats in:drawableId
!

getMatrix:arrayOf16Floats
    "get matrix"

    ^ device glxGetMatrix:arrayOf16Floats in:drawableId
!

pushMatrix
    "push down transformation matrix stack"

    ^ device glxPushmatrixIn:drawableId
!

popMatrix
    "pop transformation matrix stack"

    ^ device glxPopmatrixIn:drawableId
! !

!GLXView methodsFor:'low level transformations'!

translateX:x y:y z:z
    "translate current matrix, given individual x, y and z values."

    ^ device glxTranslateX:x y:y z:z in:drawableId
!

translateX:x 
    "translate current matrix, given an x value. Y and z are taken as 0"

    ^ device glxTranslateX:x in:drawableId
!

translateY:y 
    "translate current matrix, given a y value. X and z are taken as 0"

    ^ device glxTranslateY:y in:drawableId
!

translateZ:z
    "translate current matrix, given a z value. X and y are taken as 0"

    ^ device glxTranslateZ:z in:drawableId
!

translate:a3ElementVector
    "translate current matrix, given a float-vector."

    ^ device glxTranslate:a3ElementVector in:drawableId
!

scaleX:x y:y z:z
    "scale & mirror current matrix, given individual x, y and z values."

    ^ device glxScaleX:x y:y z:z in:drawableId
!

scale:a3ElementVector
    "scale current matrix, given a float-vector."

    ^ device glxScale:a3ElementVector in:drawableId
!

rotateX:xAngle y:yAngle z:zAngle
    "rotate the current matrix, given individual x, y and z values.
     The angles are in degrees."

    ^ device glxRotateX:xAngle y:yAngle z:zAngle in:drawableId
!

rotateIX:xAngle y:yAngle z:zAngle
    "rotate the current matrix, given individual x, y and z values.
     The angles must be integers, specifying tenths of degrees."

    ^ device glxRotateIX:xAngle y:yAngle z:zAngle in:drawableId
!

rotate:a3ElementVector
    "rotate current matrix, given a float-vector."

    ^ device glxTranslate:a3ElementVector in:drawableId
!

rotate:angle axis:axis
    "rotate the current matrix around an axis.
     The angle is in degrees.
     The axis argument must be one of the symbols #x, #y or #z"

    ^ device glxRotate:angle axis:axis in:drawableId
!

rotateX:angle
    "rotate the current matrix around x axis.
     The angle is in degrees."

    ^ device glxRotateX:angle in:drawableId
!

rotateY:angle
    "rotate the current matrix around y axis.
     The angle is in degrees."

    ^ device glxRotateY:angle in:drawableId
!

rotateZ:angle
    "rotate the current matrix around z axis.
     The angle is in degrees."

    ^ device glxRotateZ:angle in:drawableId
!

rotateIX:angle
    "rotate the current matrix around x axis.
     The angle must be integer (in tenth of degrees)."

    ^ device glxRotateIX:angle in:drawableId
!

rotateIY:angle
    "rotate the current matrix around y axis.
     The angle must be integer (in tenth of degrees)."

    ^ device glxRotateIY:angle in:drawableId
!

rotateIZ:angle
    "rotate the current matrix around z axis.
     The angle must be integer (in tenth of degrees)."

    ^ device glxRotateIZ:angle in:drawableId
! !

!GLXView methodsFor:'low level lights'!

lmdef:what index:index np:np props:props
    "define a material, light source or lighting model;
     props must be a FloatArray"

    ^ device glxLmdef:what index:index np:np props:props in:drawableId
!

lmbind:target index:index
    "select a material, light or lighting model"

    ^ device glxLmbind:target index:index in:drawableId
! !

!GLXView methodsFor:'low level textures'!

texdef2d:index nc:nc width:w height:h bits:image np:np props:props 
    ^ device glxTexDef2d:index 
		      nc:nc 
		   width:w height:h bits:image 
		      np:np props:props in:drawableId
!

tevdef:index np:np props:props
    ^ device glxTevdef:index np:np props:props in:drawableId
!

texbind:target index:index
    ^ device glxTexbind:target index:index in:drawableId
!

tevbind:target index:index
    ^ device glxTevbind:target index:index in:drawableId
! !

!GLXView methodsFor:'low level colors'!

color:index
    "set color, for non gouraud shading, we dont care if the
     argument is integer or float; otherwise, better results are
     expected with float values."

    ^ device glxColor:index in:drawableId
! 

colorRed:r green:g blue:b
    "set color, args must be integers in 0..255"

    ^ device glxColorRed:r green:g blue:b in:drawableId
!

colorRed:r green:g blue:b alpha:a
    "set color, args must be integers in 0..255"

    ^ device glxColorRed:r green:g blue:b alpha:a in:drawableId
!

mapColor:index red:red green:green blue:blue
    "assign a color to a color index"

    ^ device glxMapcolorI:index red:red green:green blue:blue in:drawableId
! !

!GLXView methodsFor:'low level polygons'!

move:a3DPoint
    "move to a position"

    ^ device glxMove:a3DPoint in:drawableId
!

moveX:x y:y z:z
    "move to a position"

    ^ device glxMoveX:x y:y z:z in:drawableId
!

drawX:x y:y z:z
    "draw to a position"

    ^ device glxDrawX:x y:y z:z in:drawableId
!

draw:a3DPoint
    "draw to a position"

    ^ device glxDraw:a3DPoint in:drawableId
!

rectangleOrigin:p1 corner:p2
    "draw a rectangle"

    ^ device glxRectX1:(p1 x) y1:(p1 y) x2:(p2 x) y2:(p2 y) in:drawableId
!

rectangleX1:x1 y1:y1 x2:x2 y2:y2
    "draw a rectangle"

    ^ device glxRectX1:x1 y1:y1 x2:x2 y2:y2 in:drawableId
!

beginPolygon 
    "start a polygon"

    ^ device glxBeginPolygonIn:drawableId
!

endPolygon 
    "end a polygon"

    ^ device glxEndPolygonIn:drawableId
! 

beginPoint 
    "start a point"

    ^ device glxBeginPointIn:drawableId
!

endPoint 
    "end a point"

    ^ device glxEndPointIn:drawableId
! 

beginClosedLine 
    "start a closed line"

    ^ device glxBeginClosedLineIn:drawableId
!

endClosedLine 
    "end a closed line"

    ^ device glxEndClosedLineIn:drawableId
! 

beginLine 
    "start a line group"

    ^ device glxBeginLineIn:drawableId
!

endLine 
    "end a line group"

    ^ device glxEndLineIn:drawableId
! 

beginTriangleMesh 
    "start a triangle mesh"

    ^ device glxBeginTriangleMeshIn:drawableId
!

endTriangleMesh 
    "end a triangle mesh"

    ^ device glxEndTriangleMeshIn:drawableId
! 

swapTriangleMesh 
    "swap triangle mesh pointers"

    ^ device glxSwaptmeshIn:drawableId
!

beginSurface 
    "start a NURBS surface def"

    ^ device glxBeginSurfaceIn:drawableId
!

endSurface
    "end a NURBS surface def"

    ^ device glxEndSurfaceIn:drawableId
! 

beginQuadrilateralStrip 
    "start a quadrilateral strip"

    ^ device glxBeginQuadrilateralStripIn:drawableId
!

endQuadrilateralStrip
    "end a quadrilateral strip"

    ^ device glxEndQuadrilateralStripIn:drawableId
! 

beginCurve 
    "start a NURBS curve def"

    ^ device glxBeginCurveIn:drawableId
!

endCurve
    "end a NURBS curve def"

    ^ device glxEndCurveIn:drawableId
!

arcX:x y:y radius:r startAngle:startang endAngle:endang
    "draw an arc"

    ^ device glxArcX:x y:y radius:r startang:startang endang:endang in:drawableId
!

arcfX:x y:y radius:r startAngle:startang endAngle:endang
    "draw a filled arc"

    ^ device glxArcX:x y:y radius:r startang:startang endang:endang in:drawableId
!

circleX:x y:y radius:r
    "draw a circle"

    ^ device glxCircX:x y:y radius:r in:drawableId
!

circlefX:x y:y radius:r
    "draw a filled circle"

    ^ device glxCircfX:x y:y radius:r in:drawableId
! !

!GLXView methodsFor:'low level spheres'!

sphWireDraw:arrayOf4Floats    
    "draw a wireframe sphere.
     The argument vector contains x,y,z and radius."

    |radius r a z step|

    device glxPushmatrixIn:drawableId.
    self translate:arrayOf4Floats.  "/ only the first 3 elements are used
    radius := arrayOf4Floats at:4.

    "
     create the latitudinal rings
    "
    0 to:180 by:20 do:[:i |
	device glxPushmatrixIn:drawableId.
	device glxRotateY:i in:drawableId.
	self arcX:0.0 y:0.0 radius:radius startAngle:0 endAngle:3599.
	device glxPopmatrixIn:drawableId.
    ].

    "
     create the longitudinal rings
    "
    device glxRotateX:90 in:drawableId. 
    -900 to:899 by:200 do:[:i |
	|t|

	t := (i * 3.14159 / 180.0).
	r := radius * t cos.
	z := radius * t sin.
	device glxPushmatrixIn:drawableId.
	device glxTranslateZ:z negated in:drawableId.
	self arcX:0.0 y:0.0 radius:r startAngle:0 endAngle:3599.
	device glxPopmatrixIn:drawableId.    
    ].                    
    device glxPopmatrixIn:drawableId.
!

sphDraw:arrayOf4Floats
    "draw a solid sphere (only on a real GL).
     The argument vector contains x,y,z and radius."

    ^ device glxSphDraw:arrayOf4Floats in:drawableId
!

sphDrawX:x y:y z:z radius:r
    "draw a solid sphere (only on a real GL)."

    ^ device glxSphDrawX:x y:y z:z radius:r in:drawableId
! !

!GLXView methodsFor:'low level patches'!

defBasis:id mat:matrix
    ^ device glxDefBasis:id mat:matrix in:drawableId
!

patchCurvesU:u v:v
    ^ device glxPatchCurvesU:u v:v in:drawableId
!

patchBasisU:u v:v
    ^ device glxPatchBasisU:u v:v in:drawableId
!

patchPrecisionU:u v:v
    ^ device glxPatchPrecisionU:u v:v in:drawableId
!

patchX:arrayOf16XFloats y:arrayOf16YFloats z:arrayOf16ZFloats
    "draw a patch"

    ^ device glxPatchX:arrayOf16XFloats 
		     y:arrayOf16YFloats 
		     z:arrayOf16ZFloats in:drawableId
! !

!GLXView methodsFor:'low level misc'!

clear
    "clear to current color"

    ^ device glxClearIn:drawableId
!

czClearColor:colorValue z:zValue
    "clear the z buffer"

    ^ device glxCzclearCval:colorValue zval:zValue in:drawableId
!

backBuffer
    "switch to backbuffer"

    ^ device glxBackBufferIn:drawableId
!

frontBuffer
    "switch to frontbuffer"

    ^ device glxFrontBufferIn:drawableId
!

swapBuffers
    "swap double buffers"

    ^ device glxSwapBuffersIn:drawableId
!

doubleBuffer
    "set double buffer mode"

    ^ device glxDoubleBufferIn:drawableId
!

singleBuffer
    "set single buffer mode"

    ^ device glxSingleBufferIn:drawableId
!

rgbMode
    "set true color mode (no colormap)"

    ^ device glxRGBmodeIn:drawableId
!

zClear
    "clear the z buffer"

    ^ device glxZClearIn:drawableId
!

zBuffer:aBoolean
    "enable/disable z buffer mode"

    ^ device glxZbuffer:aBoolean in:drawableId
!

backFace:aBoolean
    "enable/disable backface culling"

    ^ device glxBackface:aBoolean in:drawableId
!

polyFill:aBoolean
    "enable/disable polygon filling"

    ^ device glxPolyFill:aBoolean in:drawableId
!

mMode:aSymbol 
    "set the matrix mode. The argument must be one of
     #single, #viewing, #projection or #texture."

    ^ device glxMmode:aSymbol in:drawableId 
!

nMode:aSymbol 
    "set normal mode. The argument must be one of #auto or #normalize."

    ^ device glxNmode:aSymbol in:drawableId 
!

setConfig
    "must be sent after RGBmode, doubleBuffer etc. to have these
     changes really take effect. See GLX manual"

    ^ device glxGconfigIn:drawableId
! !

GLXView initialize!
