"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

PanelView subclass:#HorizontalPanelView
       instanceVariableNames:''
       classVariableNames:''
       poolDictionaries:''
       category:'Views-Layout'
!

HorizontalPanelView comment:'
COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

$Header: /claus/smalltalk/st/libwidg/RCS/HPanelV.st,v 1.8 95/02/06 00:52:25 claus Exp $
'!

!HorizontalPanelView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /claus/smalltalk/st/libwidg/RCS/HPanelV.st,v 1.8 95/02/06 00:52:25 claus Exp $
"
!

documentation
"
    a View which arranges its child-views in a horizontal row.
    All real work is done in PanelView - only the layout computation is
    redefined here.

    The layout is controlled the instance variables: 
	horizontalLayout and verticalLayout
    in addition to horizontalSpace and verticalSpace.

    The horizontal layout can be any of:

	#left           arrange elements at the left
	#leftSpace      arrange elements at the left, start with spacing
	#right          arrange elements at the right
	#rightSpace     arrange elements at the right, start with spacing
	#center         arrange elements in the center
	#spread         spread elements evenly
	#spreadSpace    spread elements evenly with spacing at the ends
	#fit            like spread, but resize elements for tight packing
	#fitSpace       like fit, with additional spacing at the far ends
	#leftFit        like left, but extend the last element to the right
	#leftSpaceFit   like leftSpace, but extend the last element to the right

    the vertical layout can be:

	#top            place element at the top
	#topSpace       place element at the top, offset by verticalSpace
	#center         place it horizontally centered
	#bottom         place it at the bottom
	#bottomSpace    place it at the bottom, offset by verticalSpace
	#fit            resize elements vertically to fit this panel
	#fitSpace       like fit, but with spacing

    The defaults is #center for both directions.
    The layout is changed by the messages #verticalLayout: and #horizontalLayout:.
    For backward compatibility (to times, where only hLayout existed), the simple
    #layout: does the same as #horizontalLayout:. Do not use this old method.

    If none of these layout/space combinations is exactly what you need in
    your application, create a subclass, and redefine the setChildPositions method.
"
!

examples
"
    These examples show the effect of various horizontalLayout and
    verticalLayout settings. Try them all. Especially, notice the
    differences between the xxx and xxxSpace layouts and the effect of
    setting different values for the spacing.
    Try resizing the view and see how the elements get rearranged.

    All of the below examples place 3 buttons onto a panel - of course,
    you can put any other view into a panel ... the last example shows this.


    example: default layout (centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	v label:'default'.

	p := HorizontalPanelView in:v.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: left-layout (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=left'.

	p horizontalLayout:#left.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: left starting with spacing (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=leftSpace'.

	p horizontalLayout:#leftSpace.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: leftFit-layout (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=leftFit'.

	p horizontalLayout:#leftFit.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: leftSpaceFit-layout (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=leftFit'.

	p horizontalLayout:#leftSpaceFit.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: right-layout (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=right'.

	p horizontalLayout:#right.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: right with initial spacing (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=rightSpace'.

	p horizontalLayout:#rightSpace.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit-layout (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	p horizontalLayout:#fit.
	v label:'hL=fit'.

	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: full fit (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	p horizontalLayout:#fit.
	p horizontalSpace:0.
	v label:'hL=fit hS=0'.

	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit with spacing (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fitSpace'.

	p horizontalLayout:#fitSpace.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: spread-layout (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	p horizontalLayout:#spread.
	v label:'hL=spread'.

	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: spread with spacing (vertical is default -> centered)

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=spreadSpace'.

	p horizontalLayout:#spreadSpace.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: spread with spacing; vertical fit

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=spreadSpace vL=fit'.

	p horizontalLayout:#spreadSpace.
	p verticalLayout:#fit.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: spread with spacing; vertical fit with spacing

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=spreadSpace vL=fitSpace'.

	p horizontalLayout:#spreadSpace.
	p verticalLayout:#fitSpace.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit - top

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fit vL=top'.

	p horizontalLayout:#fit.
	p verticalLayout:#top.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit with initial spacing

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fitSpace vL=top'.

	p horizontalLayout:#fitSpace.
	p verticalLayout:#top.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit with initial spacing in both directions

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fitSpace vL=fitSpace'.

	p horizontalLayout:#fitSpace.
	p verticalLayout:#fitSpace.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit without spacing in both directions

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fit vL=fit hS=0'.

	p horizontalLayout:#fit.
	p verticalLayout:#fit.
	p horizontalSpace:0.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit with initial spacing; top with spacing

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fitSpace vL=topSpace'.

	p horizontalLayout:#fitSpace.
	p verticalLayout:#topSpace.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit - top without spacing

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fit vL=top'.

	p horizontalLayout:#fit.
	p verticalLayout:#top.
	p horizontalSpace:0.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit - bottom with spacing and bottomSpace

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fitSpace vL=bottomSpace'.

	p horizontalLayout:#fitSpace.
	p verticalLayout:#bottomSpace.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open


    example: fit no horizontal space - bottom with spacing and bottomSpace

	|v p b1 b2 b3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=fit vL=bottomSpace'.

	p horizontalLayout:#fit.
	p verticalLayout:#bottomSpace.
	p horizontalSpace:0.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).
	b1 := Button label:'button1' in:p.
	b2 := Button label:'button2' in:p.
	b3 := Button label:'button3' in:p.
	v extent:300 @ 100.
	v open

    example: a browser like table, where the rightmost list
	     extends to the far right.

	|v p l1 l2 l3|

	v := StandardSystemView new.
	p := HorizontalPanelView in:v.
	v label:'hL=leftFit'.

	p horizontalLayout:#leftFit.
	p verticalLayout:#fit.
	p origin:(0.0 @ 0.0) corner:(1.0 @ 1.0).

	l1 := ScrollableView for:FileSelectionList in:p.
	l1 stayInDirectory:true.
	l1 ignoreParentDirectory:true.
	l1 directory:'/'.
	l1 action:[:selection | l2 directory:(l1 selectedPathname)].

	l2 := ScrollableView for:FileSelectionList in:p.
	l2 stayInDirectory:true.
	l2 directory:nil.
	l2 ignoreParentDirectory:true.
	l2 action:[:selection | l3 directory:(l2 selectedPathname)].

	l3 := ScrollableView for:FileSelectionList in:p.
	l3 directory:nil.
	l3 ignoreParentDirectory:true.
	v extent:400 @ 300.
	v open
"
! !

!HorizontalPanelView methodsFor:'accessing'!

horizontalLayout
    "return the horizontal layout as symbol.
     the returned value is one of
	#left 
	#leftSpace 
	#center
	#spread
	#fit
	#right 
	#rightSpace 
      the default is #center
    "

    ^ hLayout
!

verticalLayout
    "return the vertical layout as a symbol.
     the returned value is one of
	#top / #topSpace
	#center
	#bottom / #bottomSpace
	#fit
      the default is #center
    "

    ^ vLayout
!

horizontalLayout:aSymbol
    "change the horizontal layout as symbol.
     The argument, aSymbol must be one of:
	#left / #leftSpace 
	#center
	#spread / spredSpace
	#fit / fitSpace
	#right / #rightSpace 
      the default (if never changed) is #center
    "

    (hLayout ~~ aSymbol) ifTrue:[
	hLayout := aSymbol.
	self layoutChanged
    ]
!

verticalLayout:aSymbol
    "change the vertical layout as a symbol.
     The argument, aSymbol must be one of:
	#top / #topSpace
	#center
	#bottom / #bottomSpace
	#fit 
      the default (if never changed) is #center
    "

    (vLayout ~~ aSymbol) ifTrue:[
	vLayout := aSymbol.
	self layoutChanged
    ]
!

layout
    "leftover for historic reasons - do not use any more"

    self horizontalLayout
!

layout:aSymbol
    "leftover for historic reasons - do not use any more"

    self horizontalLayout:aSymbol
! !


!HorizontalPanelView methodsFor:'queries'!

preferedExtent
    "return a good extent, one that makes subviews fit"

    |sumOfWidths maxHeight maxWidth|

    subViews isNil ifTrue:[^ horizontalSpace @ verticalSpace].

    "compute net height needed"

    sumOfWidths := 0.
    maxHeight := 0.
    maxWidth := 0.

    subViews do:[:child |
	|childsPreference|

	childsPreference := child preferedExtent.
	sumOfWidths := sumOfWidths + childsPreference x.
	maxHeight := maxHeight max:childsPreference y.
	maxWidth := maxWidth max:childsPreference x.

"/        sumOfWidths := sumOfWidths + child widthIncludingBorder.
"/        maxHeight := maxHeight max:(child heightIncludingBorder).
"/        maxWidth := maxWidth max:(child widthIncludingBorder).
    ].
    borderWidth ~~ 0 ifTrue:[
	sumOfWidths := sumOfWidths + (horizontalSpace * 2).
	maxHeight := maxHeight + (verticalSpace * 2).
    ].
    (hLayout == #fit or:[hLayout == #fitSpace]) ifTrue:[
	sumOfWidths := maxWidth * subViews size.
	borderWidth ~~ 0 ifTrue:[
	    sumOfWidths := sumOfWidths + (verticalSpace * 2).
	]
    ] ifFalse:[
	sumOfWidths := sumOfWidths + ((subViews size - 1) * horizontalSpace).
    ].

    ((vLayout == #topSpace) or:[vLayout == #bottomSpace]) ifTrue:[
	maxHeight := maxHeight + verticalSpace
    ] ifFalse:[
	((vLayout == #fitSpace)  or:[vLayout == #center]) ifTrue:[
	    maxHeight := maxHeight + (verticalSpace * 2)
	]        
    ].

    ^ sumOfWidths @ maxHeight
! !

!HorizontalPanelView methodsFor:'layout'!

setChildPositions
    "(re)compute position of every child whenever childs are added or
     my size has changed"

    |xpos space sumOfWidths numChilds l wEach wInside|

    subViews isNil ifTrue:[^ self].

    space := horizontalSpace.
    numChilds := subViews size.
    wInside := width - (margin * 2) + (borderWidth*2) - subViews last borderWidth.

    hLayout == #fitSpace ifTrue:[
	"
	 adjust childs extents and set origins.
	 Be careful to avoid accumulation of rounding errors
	"
	wEach := (wInside - (numChilds + 1 * space)) / numChilds.
	xpos := space + margin - borderWidth.
    ] ifFalse:[
	hLayout == #fit ifTrue:[
	    "
	     adjust childs extents and set origins.
	     Be careful to avoid accumulation of rounding errors
	    "
	    wEach := (wInside - (numChilds - 1 * space)) / numChilds.
	    xpos := margin - borderWidth.
	] ifFalse:[
	    "
	     compute net width needed
	    "
	    sumOfWidths := subViews inject:0 into:[:sumSoFar :child | sumSoFar + child widthIncludingBorder].

	    l := hLayout.
	    ((l == #center) and:[numChilds == 1]) ifTrue:[
		l := #spread
	    ].
	    (l == #spread and:[numChilds == 1]) ifTrue:[
		l := #spreadSpace
	    ].

	    "
	     compute position of leftmost subview and space between them;
	     if they do hardly fit, leave no space between them 
	    "
	    (sumOfWidths >= (width - (margin * 2))) ifTrue:[
		xpos := 0.
		space := 0
	    ] ifFalse: [
		((l == #right) or:[l == #rightSpace]) ifTrue:[
		    xpos := width - (space * (numChilds - 1)) - sumOfWidths.
	"
		    borderWidth == 0 ifTrue:[
			xpos := xpos + space 
		    ].
	"
		    l == #rightSpace ifTrue:[
			xpos >= space ifTrue:[
			    xpos := xpos - space
			]
		    ].

		    xpos < 0 ifTrue:[
			space := space min:(width - sumOfWidths) // (numChilds + 1).
			xpos := width - (space * numChilds) - sumOfWidths.
		    ]
		] ifFalse:[
		    (l == #spread) ifTrue:[
			space := (width - sumOfWidths) // (numChilds - 1).
			xpos := 0.
			(space == 0) ifTrue:[
			    xpos := (width - sumOfWidths) // 2
			]
		    ] ifFalse:[
		      (l == #spreadSpace) ifTrue:[
			space := (width - sumOfWidths) // (numChilds + 1).
			xpos := space.
			(space == 0) ifTrue:[
			    xpos := (width - sumOfWidths) // 2
			]
		      ] ifFalse:[
			((l == #left) 
			or:[l == #leftSpace
			or:[l == #leftFit
			or:[l == #leftSpaceFit]]]) ifTrue:[
			    space := space min:(width - sumOfWidths) // (numChilds + 1).
			    (l == #leftSpace 
			    or:[l == #leftSpaceFit]) ifTrue:[
				xpos := space.
			    ] ifFalse:[
				xpos := 0
			    ]
	"
			    borderWidth == 0 ifTrue:[
				xpos := 0 
			    ].
	"
			] ifFalse:[
			    "center"
			    xpos := (width - (sumOfWidths
					      + ((numChilds - 1) * space))) // 2.
			    xpos < 0 ifTrue:[
				space := (width - sumOfWidths) // (numChilds + 1).
				xpos := (width - (sumOfWidths
					       + ((numChilds - 1) * space))) // 2.
			    ]
			]
		      ]
		    ]
		]
	    ].
	].
    ].

    "now set positions"

    subViews keysAndValuesDo:[:index :child |
	|ypos advance|

	vLayout == #top ifTrue:[
	    ypos := 0
	] ifFalse:[
	    vLayout == #topSpace ifTrue:[
		ypos := verticalSpace
	    ] ifFalse:[
		vLayout == #bottom ifTrue:[
		    ypos := height - child heightIncludingBorder
		] ifFalse:[
		    vLayout == #bottomSpace ifTrue:[
			ypos := height - verticalSpace - child heightIncludingBorder.
		    ] ifFalse:[
			vLayout == #fitSpace ifTrue:[
			    ypos := verticalSpace.
			    child height:(height - (verticalSpace + child borderWidth * 2))
			] ifFalse:[
			    vLayout == #fit ifTrue:[
				ypos := 0.
				child height:(height - (child borderWidth * 2))
			    ] ifFalse:[
				"centered"
				ypos := (height - child heightIncludingBorder) // 2.
			    ]
			]
		    ]
		]
	    ]
	].
	(ypos < 0) ifTrue:[ypos := 0].

	(hLayout == #fit or:[hLayout == #fitSpace]) ifTrue:[
	    child origin:(xpos truncated @ ypos)
		  corner:(xpos + wEach - (child borderWidth)) truncated
			 @ (ypos + child height).
	    advance := wEach.
	] ifFalse:[
	    child origin:(xpos @ ypos).
	    advance := child widthIncludingBorder
	].
	xpos := xpos + advance + space.

	index == numChilds ifTrue:[
	    |x|

	    hLayout == #leftFit ifTrue:[
		x := width - margin.
	    ].
	    hLayout == #leftSpaceFit ifTrue:[
		x := width - margin - space
	    ].
	    x notNil ifTrue:[
		subViews last corner:(x @ (ypos + child height))
	    ]
	]
    ].
! !
