"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

View subclass:#PanelView
       instanceVariableNames:'hLayout vLayout verticalSpace horizontalSpace mustRearrange'
       classVariableNames:''
       poolDictionaries:''
       category:'Views-Layout'
!

PanelView comment:'
COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

$Header: /claus/smalltalk/st/libwidg/RCS/PanelView.st,v 1.7 94/11/17 14:38:22 claus Exp $
'!

!PanelView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /claus/smalltalk/st/libwidg/RCS/PanelView.st,v 1.7 94/11/17 14:38:22 claus Exp $
"
!

documentation
"
    this is a view for holding subviews. (layout-widget ?!!)

    Instances of PanelView try to get all their subviews into them,
    arranging subviews left-to-right, top-to-bottom.

    If you dont like its layout, define a new subclass or use one of
    the existing subclasses: HorizontalPanelView and VerticalPanelView.

    PanelViews normally delay the actual positioning/sizing of their elements,
    until actually displayed. This is useful, if more elements are to
    be added, to avoid repeated configuration of the elements.

    If you want to query for the relative position of an element BEFORE
    the view is visible, you have to send #setChildPositionsIfChanged before
    doing so (otherwise, you may get invalid origins from the subviews). As
    an example, the modalBoxes do so before showing themselfes to ask for the
    position of the ok-button relative to the top-left box-origin, in order to
    position the ok-button under the mouse-pointer.

    Instance variables:

	hLayout         <Symbol>        controls horizontal layout; ignored in this
					class, but used in Horizontal- and
					VerticalPanelViews. See more info there.

	vLayout         <Symbol>        controls vertical layout; ignored in this
					class, but used in Horizontal- and
					VerticalPanelViews. See more info there.

	horizontalSpace <Integer>       number of pixels to use as space between elements

	verticalSpace   <Integer>       number of pixels to use as space between elements

	mustRearrange   <Boolean>       internal flag, if rearrangement is needed
"
! !

!PanelView methodsFor:'initialization'!

initialize
    super initialize.

"/    layout := #center.
    verticalSpace := ViewSpacing.
    horizontalSpace := ViewSpacing.
    mustRearrange := false
!

realize
    mustRearrange ifTrue:[
	self setChildPositions
    ].
    super realize
!

setChildPositionsIfChanged
    "set all of my child positions - this is usually delayed,
     until the panel is actually shown (since we dont know, if more
     elements are to be added) thus avoiding repositioning the elements
     over and over. However, sometimes it is nescessary, to force positioning
     the elements, for example, before querying the relative position of
     an element."

    mustRearrange ifTrue:[
	self setChildPositions
    ].

! !

!PanelView methodsFor:'accessing'!

verticalSpace:numberOfPixels
    "set the vertical space between elements (in pixels).
     The default is computed for 1mm spacing."

    verticalSpace ~= numberOfPixels ifTrue:[
	verticalSpace := numberOfPixels.
	self layoutChanged
    ]
!

horizontalSpace:numberOfPixels
    "set the horizontal space between elements on pixels (default is 1mm)"

    horizontalSpace ~= numberOfPixels ifTrue:[
	horizontalSpace := numberOfPixels.
	self layoutChanged
    ]
!

space:numberOfPixels
    "set the space between elements in pixels (default is 1mm) for both directions"

    (verticalSpace ~= numberOfPixels 
    or:[horizontalSpace ~= numberOfPixels]) ifTrue:[
	horizontalSpace := numberOfPixels.
	verticalSpace := numberOfPixels.
	self layoutChanged
    ]
!

addSubView:aView
    "redefined to recompute layout when a subview is added"

    super addSubView:aView.
    self layoutChanged
!

addSubView:newView after:aView
    "redefined to recompute layout when a subview is added"

    super addSubView:newView after:aView.
    self layoutChanged
!

addSubView:newView before:aView
    "redefined to recompute layout when a subview is added"

    super addSubView:newView before:aView.
    self layoutChanged
! !

!PanelView methodsFor:'event processing'!

sizeChanged:how
    super sizeChanged:how.
    self layoutChanged
! !

!PanelView methodsFor:'private'!

layoutChanged
    shown ifTrue:[
	self setChildPositions
    ] ifFalse:[
	mustRearrange := true
    ]
!

setChildPositions
    "(re)compute position of every child.
     This method is redefined for different layout characteristics - you may
     even create subclasses with completely different geometry management."

    |first xpos ypos maxHeightInRow thisRow|

    subViews notNil ifTrue:[
	xpos := horizontalSpace.
	ypos := verticalSpace.

	maxHeightInRow := 0.
	first := true.
	thisRow := OrderedCollection new.
	subViews do:[:child |
	    "go to next row, if this subview won't fit"
	    first ifFalse: [
		(xpos + child widthIncludingBorder + horizontalSpace) > width
		ifTrue: [
		    thisRow notEmpty ifTrue:[
			thisRow do:[:rowElement |
			    rowElement heightIncludingBorder < maxHeightInRow ifTrue:[
				rowElement top:(rowElement top + (maxHeightInRow - rowElement heightIncludingBorder))
			    ]
			]
		    ].
		    ypos := ypos + verticalSpace + maxHeightInRow.
		    xpos := horizontalSpace.
		    maxHeightInRow := 0.
		    thisRow := OrderedCollection new.
		]
	    ].
	    thisRow add:child.
	    child origin:(xpos@ypos).
	    xpos := xpos + (child widthIncludingBorder) + horizontalSpace.
	    (maxHeightInRow < (child heightIncludingBorder)) ifTrue:[
		maxHeightInRow := child heightIncludingBorder
	    ].
	    first := false
	].
	thisRow notEmpty ifTrue:[
	    thisRow do:[:rowElement |
		rowElement heightIncludingBorder < maxHeightInRow ifTrue:[
		    rowElement top:(rowElement top + (maxHeightInRow - rowElement heightIncludingBorder))
		]
	    ]
	].
    ].
    mustRearrange := false
! !
