'From Smalltalk/X, Version:2.10.4 on 24-feb-1995 at 2:29:09 am'!

TreeGraphView subclass:#WindowTreeView
	 instanceVariableNames:'showAll'
	 classVariableNames:''
	 poolDictionaries:''
	 category:'Demos-Graphics'
!

WindowTreeView comment:'
 a small application showing an ObjectView / TreeView use

 must fileIn TreeView.st, TreeGraphView.st first,
 then start with:
     WindowTreeView open

 %W% %E%
 written spring 91 by claus
'!

!WindowTreeView class methodsFor:'startup'!

open 
    |aDemo topView v|

    topView := StandardSystemView 
		    label:'Window hierarchy - known views'
		    icon:(Form fromFile:'Demo.icon' resolution:100).
    v  := HVScrollableView for:self in:topView.
    v origin:(0 @ 0) extent:(1.0 @ 1.0).

    topView open

    "WindowTreeView open"
! !

!WindowTreeView methodsFor:'generating the class-tree picture'!

addChildrenToTreeFrom:anElement father:father xPos:xPos yPos:yPos first:isFirst
    ^ super addChildrenToTreeFrom:anElement father:father xPos:xPos yPos:yPos first:false
!

childrenOf:aView do:aBlock
    |subviews|

    (aView respondsTo:#subViews) ifFalse:[
	"not a real view"

	^ self
    ].

    showAll ifFalse:[
	"dont show ShadowViews, popups etc."

	aView createOnTop ifTrue:[
	    ^ self
	]
    ].

    subviews := aView subViews.
    subviews notNil ifTrue:[
	subviews do:aBlock
    ]
!

textFor:aView
    (aView isKindOf:StandardSystemView) ifTrue:[
	^ (aView class name , ' (' , aView label , ')')
    ].
    ^ aView class name
!

addChildrenToTreeFrom:aView xPos:xPos yPos:yPos
    "special - we have a forest here ..."

    |newObject yNext yPrev|

    yNext := yPos.
    aView isNil ifTrue:[
	Display allViewsDo:[:v |
	    v superView isNil ifTrue:[
		"its a top-view"

		(v respondsTo:#subViews) ifTrue:[
		    "a real view, dont show if its a modal or
		     popup view, except when currently visible"
		    (showAll or:[v realized or:[v createOnTop not]]) ifTrue:[
			"dont show ShadowViews, popups etc."
			yPrev := yNext.
			yNext := self addChildrenToTreeFrom:v father:nil xPos:xPos yPos:yNext first:false.

			"anything added ?"
			yNext ~~ yPrev ifTrue:[
			    "create a gap"
			     yNext := yNext + (device verticalPixelPerMillimeter*10) rounded.
			]
		    ]
		]
	    ]
	]
    ].
    ^ yNext
!

colorFor:aView
    (aView realized and:[
     aView topView realized and:[
     aView topView shown]]) ifTrue:[
	device hasGreyscales ifTrue:[
	    ^ Color red
	].
	^ Color veryDarkGrey
    ].
    ^ Black
! !

!WindowTreeView methodsFor:'initialization'!

middleButtonMenu
    |labels selectors m v org corner ext|

    selection isNil ifTrue:[
	labels := #(
			    'update'
		   ).

	selectors := #(
			    doUpdate
		      ).

	^ PopUpMenu  labels:(resources array:labels)
		     selectors:selectors
		     receiver:self
    ].

    labels := #(
			'update'
			'-'
			'inspect'
			'bounds ...'
			'browse class'
			'-'
			'raise'
			'lower'
			'-'
			'destroy'
		).

    selectors := #(
			doUpdate
			nil
			doInspect
			boundsMenu
			doBrowse
			nil
			doRaise
			doLower
			nil
			doDestroy
		  ).

    m := PopUpMenu  labels:(resources array:labels)
		 selectors:selectors
		 receiver:self.

    v := (assoc at:selection).
    v relativeOrigin notNil ifTrue:[
	org := v relativeOrigin
    ] ifFalse:[
	org := v origin
    ].
    v relativeCorner notNil ifTrue:[
	corner := v relativeCorner
    ] ifFalse:[
	corner := v corner
    ].
    v relativeExtent notNil ifTrue:[
	ext := v relativeExtent
    ] ifFalse:[
	ext := v extent
    ].

    m subMenuAt:#boundsMenu
	    put:(PopUpMenu
			labels:(Array 
				    with:('origin:' , org printString)
				    with:('corner:' , corner printString)
				    with:('extent:' , ext printString)
			       )
			selectors:nil
			receiver:nil).
    ^ m
!

initializeMiddleButtonMenu
    self model:self.
    self menu:#middleButtonMenu
!

initialize
    topNode := nil.
    showAll := false.
    super initialize.
! !

!WindowTreeView methodsFor:'user interaction'!

doRaise
    self withSelectedElementDo:[:theView |
	 theView "topView "raise
    ]
!

doUpdate
    self updateTree.
    self contentsChanged
!

doInspect
    self withSelectedElementDo:[:theView |
	 theView inspect
    ]
!

doBrowse
    self withSelectedElementDo:[:theView |
	 SystemBrowser browseClass:(theView class)
    ]
!

doLower
    self withSelectedElementDo:[:theView |
	 theView topView lower
    ]
!

doDestroy
    self withSelectedElementDo:[:theView |
	theView destroy.
	self updateTree
    ]
! !

